/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

#include "appConnCli.h"

/*---------------------------------------------------------------------------*/

/* Include header files for new connections here */
#include "conn1988Cli.h"
#include "connV2Cli.h"

/*---------------------------------------------------------------------------*/

void 
appConnClientInit(void)
{
  appConnType* ai;
     
  /* Initialize the network layer */
  nConnTypesInit();

  /* 1988 */
  ai = findAppConnTypeNamed("Z39501988");
  if (ai == NULL)
   { ai = (appConnType*)safeMalloc(sizeof(appConnType));
     ai->name = safeStrdup("Z39501988");
   }
  ai->appConnOpenFunc = z39501988Open;
  ai->appConnFreeFunc = z39501988Free;
  ai->pduInfoFunc = z39501988PduInfo;
  ai->initReqWriteFunc = z39501988InitReqWrite;
  ai->initRespReadFunc = z39501988InitRespRead;
  ai->searchReqWriteFunc = z39501988SearchReqWrite;
  ai->searchRespReadFunc = z39501988SearchRespRead;
  ai->nextHeadlinesReqWriteFunc = z39501988NextHeadlinesReqWrite;
  ai->nextHeadlinesRespReadFunc = z39501988NextHeadlinesRespRead;
  ai->retrieveReqWriteFunc = z39501988RetrieveReqWrite;
  ai->retrieveRespReadFunc = z39501988RetrieveRespRead;
  registerAppConnType(ai);

  /* V2 */
  ai = findAppConnTypeNamed("Z3950V2");
  if (ai == NULL)
   { ai = (appConnType*)safeMalloc(sizeof(appConnType));
     ai->name = safeStrdup("Z3950V2");
   }
  ai->appConnOpenFunc = z3950V2Open;
  ai->appConnFreeFunc = z3950V2Free;
  ai->pduInfoFunc = z3950V2CliPduInfo;
  ai->initReqWriteFunc = z3950V2InitReqWrite;
  ai->initRespReadFunc = z3950V2InitRespRead;
  ai->searchReqWriteFunc = z3950V2SearchReqWrite;
  ai->searchRespReadFunc = z3950V2SearchRespRead;
  ai->nextHeadlinesReqWriteFunc = z3950V2NextHeadlinesReqWrite;
  ai->nextHeadlinesRespReadFunc = z3950V2NextHeadlinesRespRead;
  ai->retrieveReqWriteFunc = z3950V2RetrieveReqWrite;
  ai->retrieveRespReadFunc = z3950V2RetrieveRespRead;
  registerAppConnType(ai);
}

/*---------------------------------------------------------------------------*/

appConnStatus 
appConnInitReqWrite(appConn* conn,
		    char*    user,    
		    char*    password,
		    char*    clientInfo,
		    any*     userInfo,
		    list*    diags)
{
  appConnType* ai;

  if (conn == NULL || 
      (ai = findAppConnTypeNamed(conn->name)) == NULL ||
      ai->initReqWriteFunc == NULL)
    return(appConnStatusUndefinedFunction);
  else
   { appConnStatus res;
     res = (*ai->initReqWriteFunc)(conn,user,password,clientInfo,
				   userInfo,diags);
     return(res);
   }
}

/*---------------------------------------------------------------------------*/

appConnStatus 
appConnInitRespRead(appConn* conn,  
		    any**    userInfo,
		    list*    diags)
{
  appConnType* ai;

  if (conn == NULL || 
      (ai = findAppConnTypeNamed(conn->name)) == NULL ||
      ai->initRespReadFunc == NULL)
    return(appConnStatusUndefinedFunction);
  else
   { appConnStatus res;
     res = (*ai->initRespReadFunc)(conn,userInfo,diags);
     return(res);
   }
}

/*---------------------------------------------------------------------------*/

appConnStatus 
appConnSearchReqWrite(appConn*     conn,  
		      long         maxHits,
		      long         hitsRequested, 
		      long         rankingScheme, 
		      char*        dbName, 
		      appConnQry*  qry,
		      any*         userInfo,
		      list*        diags)
{
  appConnType* ai;

  if ((maxHits < 0L) || hitsRequested < 0L)
    return(appConnStatusBadArguments);

  if (conn == NULL || 
      (ai = findAppConnTypeNamed(conn->name)) == NULL ||
      ai->searchReqWriteFunc == NULL)
    return(appConnStatusUndefinedFunction);
  else
   { appConnStatus res;
     res = (*ai->searchReqWriteFunc)(conn,maxHits,hitsRequested, 
				     rankingScheme,dbName,qry,
				     userInfo,diags);
     return(res);
   }
}

/*---------------------------------------------------------------------------*/

appConnStatus 
appConnSearchRespRead(appConn* conn,
		      long*    totalHits, 
		      list*    qryTermsUsed,
		      list*    hits,
		      any**    userInfo,
		      list*    diags)
{
  appConnType* ai;

  if (conn == NULL || 
      (ai = findAppConnTypeNamed(conn->name)) == NULL ||
      ai->searchRespReadFunc == NULL)
    return(appConnStatusUndefinedFunction);
  else
   { appConnStatus res;
     res = (*ai->searchRespReadFunc)(conn,totalHits,qryTermsUsed,
				     hits,userInfo,diags);
     return(res);
   }
}

/*---------------------------------------------------------------------------*/

appConnStatus 
appConnNextHeadlinesReqWrite(appConn* conn,
			     long     startingFromHit, 
			     long     hitsRequested,
			     any*     userInfo,
			     list*    diags)
{
  appConnType* ai;

  if (conn == NULL || 
      (ai = findAppConnTypeNamed(conn->name)) == NULL ||
      ai->nextHeadlinesReqWriteFunc == NULL)
    return(appConnStatusUndefinedFunction);
  else
   { appConnStatus res;
     res = (*ai->nextHeadlinesReqWriteFunc)(conn,startingFromHit,
					    hitsRequested,userInfo,diags);
     return(res);
   }
}

/*---------------------------------------------------------------------------*/

appConnStatus 
appConnNextHeadlinesRespRead(appConn* conn,
			     list*    hits,
			     any**    userInfo,
			     list*    diags)
{
  appConnType* ai;

  if (conn == NULL || 
      (ai = findAppConnTypeNamed(conn->name)) == NULL ||
      ai->nextHeadlinesRespReadFunc == NULL)
    return(appConnStatusUndefinedFunction);
  else
   { appConnStatus res;
     res = (*ai->nextHeadlinesRespReadFunc)(conn,hits,userInfo,diags);
     return(res);
   }
}

/*---------------------------------------------------------------------------*/

appConnStatus 
appConnRetrieveReqWrite(appConn* conn,
			long     retrievalType,
			long     resSetPos, 
			any*     recId, 
			char*    docId, 
			char*    elemId, 
			char*    varId, 
			char*    dbName,
			long     startByte,
			long     lengthInBytes,
			any*     userInfo,
			list*    diags)
{
  appConnType* ai;

  if (conn == NULL || 
      (ai = findAppConnTypeNamed(conn->name)) == NULL ||
      ai->retrieveReqWriteFunc == NULL)
    return(appConnStatusUndefinedFunction);
  else
   { appConnStatus res;
     res = (*ai->retrieveReqWriteFunc)(conn,retrievalType,resSetPos, 
				       recId,docId,elemId,varId,dbName,
				       startByte,lengthInBytes,
				       userInfo,diags);
     return(res);
   }
}

/*---------------------------------------------------------------------------*/

appConnStatus 
appConnRetrieveRespRead(appConn* conn, 
			char**   data,
			long*    dataLength,
			any**    userInfo,
			list*    diags)
{ 
  appConnType* ai;

  if (conn == NULL || 
      (ai = findAppConnTypeNamed(conn->name)) == NULL ||
      ai->retrieveRespReadFunc == NULL)
    return(appConnStatusUndefinedFunction);
  else
   { appConnStatus res;
     res = (*ai->retrieveRespReadFunc)(conn,data,dataLength,userInfo,diags);
     return(res);
   }
}

/*-------------------------------------------------------------------------- */


