/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

#define cutil_c

#include <ctype.h>
#include <string.h>
#include <time.h>

#include "cutil.h"
#include "panic.h"
#include "futil.h"

long myPID = -1;

typedef long (longfunc)(long c);

#define MALLOC_DEBUG
long md = 0;

/*---------------------------------------------------------------------------*/
/* Math utilities */
/*---------------------------------------------------------------------------*/

boolean
isValidBigNum(char* word)
/* see if the word has any hope of being a bignum */
{
  long i;
  for (i = 0; word[i] != '\0'; i++)
   { if (!isdigit(word[i]))
       return(false);
   }
  return(true);
}

/*---------------------------------------------------------------------------*/

boolean 
bigNumEQ(char* v1,char* v2)
/* given 2 char strings representing numbers (eg "2354123"), compare them.
   NOTE - only possitive integers handled for now 
 */
{
  if (v1 == NULL || v2 == NULL)
    return(false); /* numbers can't be NULL */
  else
    return(strcmp(v1,v2) == 0);
}

/*---------------------------------------------------------------------------*/

boolean 
bigNumLE(char* v1,char* v2)
/* given 2 char strings representing numbers (eg "2354123"), compare them.
   NOTE - only possitive integers handled for now (no leading 0's either)
 */
{
  if (v1 == NULL || v2 == NULL)
    return(false); /* numbers can't be NULL */
  else
   { long s1 = strlen(v1);
     long s2 = strlen(v2);
     if (s1 > s2)
       return(false);
     if (s1 < s2)
       return(true);
     else
      { long i;
	for (i = 0; i < s1; i++)
	 { if (v1[i] > v2[i])
	     return(false);
	   if (v1[i] < v2[i])
	     return(true);
	 }
      }
     /* if we get here they are equal */
     return(true);
   }
}

/*---------------------------------------------------------------------------*/

boolean 
bigNumLT(char* v1,char* v2)
/* given 2 char strings representing numbers (eg "2354123"), compare them.
   NOTE - only possitive integers handled for now 
 */
{
  return(bigNumEQ(v1,v2) == false && bigNumLE(v1,v2));
}

/*---------------------------------------------------------------------------*/

boolean 
bigNumGE(char* v1,char* v2)
/* given 2 char strings representing numbers (eg "2354123"), compare them.
   NOTE - only possitive integers handled for now 
 */
{
  return(bigNumLE(v2,v1));
}

/*---------------------------------------------------------------------------*/

boolean 
bigNumGT(char* v1,char* v2)
/* given 2 char strings representing numbers (eg "2354123"), compare them.
   NOTE - only possitive integers handled for now 
 */
{
  return(bigNumEQ(v1,v2) == false && bigNumGE(v1,v2));
}

/*---------------------------------------------------------------------------*/

long
bigNumCmp(void** v1,void** v2)
/* arguments are pointers to strings containing bigNums, 
   suitable for use with qsort. sorts in numeric order
 */
{
  if (bigNumEQ(*(char**)v1,*(char**)v2))
    return(0);
  else if (bigNumLE(*(char**)v1,*(char**)v2))
    return(-1);
  else if (bigNumGE(*(char**)v1,*(char**)v2))
    return(1);
  else /* something is wrong! the arguments were not bignums! */
    return(0);
}

/*---------------------------------------------------------------------------*/

long
longCmp(void** s1,void** s2)
/* Arguments are pointers to pointers to longs, suitable for use with qsort.
   They're pointers to pointers to long integers, because qsort
   passes in the *address* of the two elements to compare.
 */
{
  return(*(*(long**)s1) - *(*(long**)s2));
}

/*---------------------------------------------------------------------------*/
/* String utilities */
/*---------------------------------------------------------------------------*/

char*
doSafeStrncat(char* dst,char* src,size_t maxToAdd,size_t maxTotal)
{
  size_t dstSize = strlen(dst);
  size_t srcSize = strlen(src);
  
  if (dstSize + srcSize < maxTotal) 
    return(strncat(dst,src,maxToAdd));
  else
    return(strncat(dst,src,MIN(maxToAdd, maxTotal - dstSize - 1)));
    
/****
  Ben fixed a bug here, 2/8/94.  we should not assume that src[] can be
  modified, because it may be a string constant, e.g. " ";  some machines
  crash when you try to modify their program space.
  {
     size_t truncateTo = maxTotal - dstSize - 1;
     char   saveChar;
     char*  result = NULL;

     saveChar = src[truncateTo];
     src[truncateTo] = '\0';
     result = strncat(dst,src,maxToAdd);
     src[truncateTo] = saveChar;
     return(result);
   }
***/

}

/*---------------------------------------------------------------------------*/

char*
doSafeStrncpy(char* s1,char* s2,long n)
{
  s1[n-1] = '\0';
  return(strncpy(s1,s2,n - 1));
}

/*---------------------------------------------------------------------------*/

long
safeStrlen(char* s)
{
  if (s == NULL)
    return(0L);
  else
    return(strlen(s));
}

/*---------------------------------------------------------------------------*/

char*
safeStrdup(char* s)
{
  unsigned long len;
  char* copy = NULL;
  
  if (s == NULL)		
    return(NULL);
    
  len = strlen(s);		
  copy = (char*)safeMalloc((size_t)(sizeof(char)*(len + 1)));
  strncpy(copy,s,len + 1);
  return(copy);
}

/*---------------------------------------------------------------------------*/

char*
safeStrndup(char* s,long n)
{
  char* copy = NULL;
  
  if (s == NULL)
    return(NULL);
    
  copy = (char*)safeMalloc((size_t)(sizeof(char)*n));
  copy[n-1] = '\0';
  strncpy(copy,s,n-1);
  return(copy);
}

/*---------------------------------------------------------------------------*/

char* 	
safeStrAppend(char* s,char* ap)
/* make a new string by appending s and ap, and return the new string.
   frees s.
 */
{
  char* ns = (char*)safeMalloc(safeStrlen(s) + safeStrlen(ap) + 1);
  if (s != NULL)
    strcpy(ns,s);
  if (ap != NULL)
    strcat(ns,ap);
  safeFree(s);
  return(ns);
}

/*---------------------------------------------------------------------------*/

char*
strtokf(char* s1,longfunc* isDelimiter)
{
  static char* searchStr = NULL;
  static longfunc* delimiterFunc;
  long i;
  char* startTok = NULL;

  if (s1 != NULL)		
    searchStr = s1;
    
  if (isDelimiter != NULL)
    delimiterFunc = isDelimiter;
   
  if (searchStr == NULL || searchStr[0] == '\0')
    return(NULL);		
    
  if (delimiterFunc == NULL)
    return(NULL);		
  
  for (i = 0; searchStr[i] != '\0'; i++)
   { if ((*delimiterFunc)((long)searchStr[i]) == false) /* not delimiter */
       break;
   }
  
  if (searchStr[i] == '\0') 
    return(NULL);		
  else
    startTok = searchStr + i;	

  for (; searchStr[i] != '\0'; i++)
   { if ((*delimiterFunc)((long)searchStr[i]) == true) /* is delimiter */
       break;
   }
  
  if (searchStr[i] != '\0')
   { searchStr[i] = '\0';
     searchStr = searchStr + i + 1;
   }
  else
    searchStr = searchStr + i;
   
  return(startTok);
}

/*---------------------------------------------------------------------------*/

char*
strtokfIsalnum(char* s1)
{
  static char* searchStr = NULL;
  long i;
  char* startTok = NULL;

  if (s1 != NULL)		
    searchStr = s1;

  if (searchStr == NULL || searchStr[0] == '\0')
    return(NULL);		
  
  for (i = 0; searchStr[i] != '\0'; i++)
   { if (safeIsAlNum(searchStr[i]))
       break;
   }
  
  if (searchStr[i] == '\0') 
    return(NULL);		
  else
    startTok = searchStr + i;	
  
  for (; searchStr[i] != '\0'; i++)
   { if (!safeIsAlNum(searchStr[i]))
       break;
   }
   
  if (searchStr[i] != '\0')
   { searchStr[i] = '\0';
     searchStr = searchStr + i + 1;
   }
  else
    searchStr = searchStr + i;
   
  return(startTok);
}

/*---------------------------------------------------------------------------*/

boolean 
substrcmp(char* string1,char* string2)
/* return true if they match up to the length of the smaller (but not if the
   smaller is empty)
 */
{
  
  register char* a,*b;
  
  a = string1;
  b = string2;
  
  /* see if they are both empty - if so them match */
  if (*a == '\0' && *b == '\0')
    return(true);
  
  /* see if either is empty (but not both) - no match */
  if (*a == '\0' || *b == '\0')
    return(false);

  /* otherwise we need to look through each byte */
  while (*a && *b)  
   { if (*a++ != *b++) 
       return(false);
   }
  return(true);
}

/*---------------------------------------------------------------------------*/

char 
charDowncase(unsigned long longCh)
{
  unsigned char ch = longCh & 0xFF; 
  
  return(((ch >= 'A') && (ch <= 'Z')) ? (ch + 'a' -'A') : ch);
}

/*---------------------------------------------------------------------------*/

char*
stringDowncase(char* word)
{
  long i = 0;

  if (word == NULL)
    return(NULL);

  while (word[i] != '\0')
   { word[i] = charDowncase((unsigned long)word[i]);
     i++;
   }

  return(word);
}

/*---------------------------------------------------------------------------*/

long
alphaCmp(void** s1,void** s2)
/* arguments are pointers to strings, suitable for use with qsort.
   sorts in alphabetical order 
 */
{
  return((long)strcmp(*(char**)s1,*(char**)s2));
}

/*---------------------------------------------------------------------------*/

long
revAlphaCmp(void** s1,void** s2)
/* arguments are pointers to strings, suitable for use with qsort.
   sorts in reverse alphabetical order 
 */
{
  return((long)strcmp(*(char**)s1,*(char**)s2));
}

/*---------------------------------------------------------------------------*/

long
longStrCmp(void** s1,void** s2)
/* arguments are pointers to strings containing the ascii representation of
   long intergers, suitable for use with qsort.
   sorts in numeric order.
 */
{
  long l1,l2;
  
  l1 = atol(*(char**)s1);
  l2 = atol(*(char**)s2);
  
  return(l1 - l2);
}

/*---------------------------------------------------------------------------*/

long
revLongStrCmp(void** s1,void** s2)
/* arguments are pointers to strings containing the ascii representation of
   long intergers, suitable for use with qsort.
   sorts in reverse numeric order.
 */
{
  long l1,l2;
  
  l1 = atol(*(char**)s1);
  l2 = atol(*(char**)s2);
  
  return(l2 - l1);
}

/*---------------------------------------------------------------------------*//* Heap utilities */
/*---------------------------------------------------------------------------*/

void
checkPtr(void* ptr,char* msg,long val)
/* check the ptr to see if it is a ptr to valid memory, and error if not */
{ 
  if (ptr == NULL)
   { panic(
"%s %ld bytes, consider increasing virtual memory, or running fewer programs",
	   msg,val);
   }
}

/*---------------------------------------------------------------------------*/

void*
doSafeMalloc(size_t size)
/* allocates size bytes and clears them (sets them to 0) */
{ 
  void* ptr = NULL;

#ifdef MALLOC_DEBUG
  if (md > 0)
   { fprintf(stderr,"safeMalloc of %ld bytes ",size);
     fflush(stderr);
   }
#endif

  if (size <= 0) 
   { 
#ifdef MALLOC_DEBUG
     if (md > 0)
       fprintf(stderr,"returns NULL\n");
#endif
     return(NULL);
   }

#ifdef THINK_C
  ptr = (void*)NewPtr((long)size);
  checkPtr(ptr,"unable to allocate",(long)size);
  memset(ptr,0L,(size_t)size);	
#else  
  ptr = (void*)calloc((size_t)size,(size_t)1);
  checkPtr(ptr,"unable to allocate",(long)size);
#endif

#ifdef MALLOC_DEBUG
  if (md > 0)
    fprintf(stderr,"returns %lx\n",ptr);
#endif
  
  return(ptr);
}

/*---------------------------------------------------------------------------*/

void*
doSafeRealloc(void* ptr,size_t size)
/* note this does not guarantee setting any newly allocated bytes to 0 */
{ 
  register void* nptr = NULL;
  
#ifdef MALLOC_DEBUG
  if (md > 0)
   { fprintf(stderr,"safeRealloc %ld bytes on %lx ",size,ptr);
     fflush(stderr);
   }
#endif

  if (size <= 0)
   {
#ifdef MALLOC_DEBUG
     if (md > 0)
       fprintf(stderr,"returns NULL\n");
#endif
     return(NULL);
   }

  if (ptr == NULL)		
    return(safeMalloc(size));
    
#ifdef THINK_C
  nptr = NewPtr(size);		 
  checkPtr(nptr,"unable to reallocate",(long)size);
  BlockMove(ptr,nptr,size);	
  DisposPtr(ptr);		    
#else
  nptr = (void*)realloc(ptr,size);
  checkPtr(nptr,"unable to reallocate",(long)size);
#endif  
   
#ifdef MALLOC_DEBUG
  if (md > 0)
    fprintf(stderr,"returns %lx\n",nptr);
#endif
  
  return(nptr);
}

/*---------------------------------------------------------------------------*/

/*
void* watchPtr = (void*)-1;
*/

void
doSafeFree(void* ptr)
/* free memory allocated with safeMalloc or safeRealloc */
{

#ifdef MALLOC_DEBUG
  if (md > 0)
    fprintf(stderr,"safeFree %lx\n",ptr);
#endif

/*  
  if (ptr == watchPtr)
    pauseForUser("freeing watchPtr");
*/

  if (ptr != NULL)		
   {				
#ifdef THINK_C
     DisposPtr(ptr);
#else
     free(ptr);
#endif
   }
}

/*---------------------------------------------------------------------------*/
/* Memory Block Utilities */
/*---------------------------------------------------------------------------*/

any* 
makeAny(unsigned long size,char* data)
{
  any* a = (any*)safeMalloc((size_t)sizeof(any));
  a->size = size;
  a->bytes = data;
  return(a);
}

/*----------------------------------------------------------------------*/

void
freeAny(any* a)
{
  if (a != NULL)
   { if (a->bytes != NULL)
       safeFree(a->bytes);
     safeFree(a);
   }
}

/*----------------------------------------------------------------------*/

any* 
duplicateAny(any* a)
{
  any* copy = NULL;

  if (a == NULL)
    return(NULL);

  copy = (any*)safeMalloc((size_t)sizeof(any));
  copy->size = a->size;
  if (a->bytes == NULL)
    copy->bytes = NULL;
  else
   { copy->bytes = (char*)safeMalloc((size_t)copy->size);
     memcpy(copy->bytes,a->bytes,(size_t)copy->size);
   }
  return(copy);
}

/*----------------------------------------------------------------------*/

any*
stringToAny(char* s)
{
  any* a = NULL;
  
  if (s == NULL)
    return(NULL);
    
  a = (any*)safeMalloc((size_t)sizeof(any));
  a->size = strlen(s);
  a->bytes = (char*)safeMalloc((size_t)a->size);
  memcpy(a->bytes,s,(size_t)a->size);
  return(a);
}

/*----------------------------------------------------------------------*/

char*
anyToString(any* a)
{
  char* s = NULL;
  
  if (a == NULL)
    return(NULL);
    
  s = safeMalloc((size_t)(a->size + 1));
  memcpy(s,a->bytes,(size_t)a->size);
  s[a->size] = '\0';
  return(s);
}

/*----------------------------------------------------------------------*/

any* 
longToAny(long num)
{
  char s[40];

  sprintf(s,"%ld",num);
  
  return(stringToAny(s));
}

/*----------------------------------------------------------------------*/

long
anyToLong(any* a)
{
  long num;
  char* str = NULL;
  str = anyToString(a);
  sscanf(str,"%ld",&num);	
  safeFree(str);
  return(num);
}
 
/*----------------------------------------------------------------------*/

boolean
anycmp(any* a1,any* a2)
{
  if ((a1 == NULL) || (a2 == NULL))
    return(false);
    
  if ((a1->size == a2->size) &&
      !(strncmp(a1->bytes,a2->bytes,a1->size)))
    return(true);
  else return(false);
}

/*---------------------------------------------------------------------------*/
/* Miscelaneous Utilities */
/*---------------------------------------------------------------------------*/

void
initMyPID(void)
{
  if (myPID == -1)
    myPID = getpid();
}

/*---------------------------------------------------------------------------*/

void
pauseForUser(char* message)
{
#ifdef THINK_C
  Debugger();
#else
  printf("%s\n<press return to continue>\n",message);
  getchar();
#endif
}

/*---------------------------------------------------------------------------*/

char*
nextArg(long* argc,char*** argv)
{
  if ((*argc)-- > 0)
   { char* arg = *((*argv)++);
     return(arg);
   }
  else
   { return(NULL);
   }
}

/*---------------------------------------------------------------------------*/

char* 
peekArg(long* argc,char*** argv)
{
  if ((*argc) > 0)
    return(**argv);
  else
    return(NULL);
}

/*---------------------------------------------------------------------------*/

char*
readNextArg(long* argc,char*** argv,boolean* useFile,FILE* file)
/* 
   Like nextArg(),but can read arguments from a file (one line per arg).
   Note you must copy the returned value if you need access to more
   than one argument at a time.
 */
{
  static char buf[300];

  if (file != NULL && useFile != NULL && *useFile)
   { if (safeFGets(buf,300,file) == NULL) /* nothing more on the file */
      { *useFile = false;
	return(nextArg(argc,argv)); /* start reading regular args again */
      }
     else /* return a line from the file */
      { return(buf);
      }
   }
  else /* just read regular args from the command line */
   { return(nextArg(argc,argv));
   }
}

/*---------------------------------------------------------------------------*/

#ifdef THINK_C

#if THINK_C == 1
#include <EventMgr.h>
#endif 

#if THINK_C == 5
#include <Events.h>
#endif

#ifdef WAIStation
#include "CRetrievalApp.h"
#endif 
#endif 

void
beFriendly(void)

{ 
#ifdef never
#ifdef THINK_C
   EventRecord  	macEvent;	

   static RgnHandle	mouseRgn = NULL; 
   long		        sleepTime; 	
   
#ifdef WAIStation
   gApplication->FrobWaitCursor();
#endif 
   
   if (mouseRgn == NULL)
     mouseRgn = NewRgn(); 
     
   sleepTime = 5; 
   
   WaitNextEvent(everyEvent,&macEvent,sleepTime,mouseRgn); 
#endif 
#endif
}

/*---------------------------------------------------------------------------*/
/* Printing utilities */
/*---------------------------------------------------------------------------*/

long
cprintf(boolean print,char* format,...)
{
  va_list ap;			
  if (print == 1)
    { long res;
      va_start(ap,format);	
      res = vprintf(format,ap);	
      va_end(ap);		
      return(res);
    }
  else
    return(0);
}

/*---------------------------------------------------------------------------*/

void
printHexByte(unsigned char h) 
{
  if (h < 0x10)
    printf("0%x",h);
  else
    printf("%x",h);
}

/*---------------------------------------------------------------------------*/

char*
commafy(long val,char* buf)
/* 
   prints the value using format, then adds commas separating every
   three orders of magnitude on the left side of the decimal point.

   limitations
   - there needs to be a buffer to write the result into.  We provide
     one for convenience, but of course that's not reentrant, so if you
     are calling this is multiple times before using the result you'll
     need to pass in your own buffer
   - only works on longs
   - only works on positive nums
   - no other formatting allowed
*/
{
  static char myBuf[100];

  if (buf == NULL)
    buf = myBuf;

  if (val < 1000) /* 1 thousand */
    sprintf(buf,"%ld",val);
  else if (val < 1000000) /* 1 millon */
    { long lessThanAGrand = val%1000;
      long moreThanAGrand = val - lessThanAGrand;
      sprintf(buf,"%ld,%03ld",moreThanAGrand/1000,lessThanAGrand);
    }
  else if (val < 1000000000) /* 1 trillion, 1 gig */
    { long lessThanAGrand = val%1000;
      long lessThanAMil = (val%1000000) - lessThanAGrand;
      long moreThanAMil = val - lessThanAMil - lessThanAGrand;
      sprintf(buf,"%ld,%03ld,%03ld",moreThanAMil/1000000,lessThanAMil/1000,
	      val%1000);
    }

  return(buf);
}

/*---------------------------------------------------------------------------*/

void
printTime(FILE* f,time_t t)
{
  if (f == NULL)
    f = stdout;
  
  fprintf(f,"%s",printableTime(t));
}

/*---------------------------------------------------------------------------*/

char*
printableTime(time_t t)
{ 
  static char* string;

  string = ctime(&t);

  if (string)
   { if (string[strlen(string)-1] == '\n')
       string[strlen(string)-1] = '\0';   
     return(string+4);
   }
  else
    return("Time Unknown");
}

/*---------------------------------------------------------------------------*/

char*
printableTimeNow(void)
{ 
  time_t t;
  time(&t);
  return(printableTime(t));
}

/*---------------------------------------------------------------------------*/
/* Timer utilities */
/*---------------------------------------------------------------------------*/

typedef struct milliTime
{ time_t timeInSec;
  long   milliSec;
} milliTime;

/* global variables - yuck - needed to communicate between startTimer() and
   stopTimer()
 */
#define MAX_TIMERS 10
milliTime startTime[MAX_TIMERS];
milliTime stopTime[MAX_TIMERS];

#if (defined(bsdi) || defined(__svr4__))
/* do nothing, these OS's don't support millisecond time */
#else
#include <sys/timeb.h>
#endif

boolean 
startTimer(long timerNum)
/* start timerNum */
{
  if (timerNum < MAX_TIMERS)
   { 
#if (defined(bsdi) || defined(__svr4__))
     /* these OS's can't give millisecond time */
     startTime[timerNum].timeInSec = time(NULL);
     startTime[timerNum].milliSec = 0;
#else
     static struct timeb t;
     ftime(&t);
     startTime[timerNum].timeInSec = t.time;
     startTime[timerNum].milliSec = t.millitm;
#endif
     return(true);
   }
  return(false);
}

/*---------------------------------------------------------------------------*/

boolean 
stopTimer(long timerNum)
/* stop timerNum */
{
  if (timerNum < MAX_TIMERS)
   { 
#if (defined(bsdi) || defined(__svr4__))
     /* these OS's can't give millisecond time */
     stopTime[timerNum].timeInSec = time(NULL);
     stopTime[timerNum].milliSec = 0;
#else
     static struct timeb t;
     ftime(&t);
     stopTime[timerNum].timeInSec = t.time;
     stopTime[timerNum].milliSec = t.millitm;
#endif
     return(true);
   }
  return(false);
}

/*---------------------------------------------------------------------------*/

long
timerVal(long timerNum)
/* return elapsed time for timerNum in millisec */
{
  if (timerNum < MAX_TIMERS)
   { return((1000 * (stopTime[timerNum].timeInSec - 
		     startTime[timerNum].timeInSec)) +
	     (stopTime[timerNum].milliSec - startTime[timerNum].milliSec));
   }
  else
    return(-1);
}

/*---------------------------------------------------------------------------*/

