/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

#ifndef cutil_h
#define cutil_h

#include <stdarg.h>
#include <stdio.h>   
#include <math.h> /* for min/maxlong, may need <values.h> too  */
#ifndef EXIT_SUCCESS 
#ifndef bsdi
#include <stdlib.h> 
#endif
#endif /* ndef EXIT_SUCCESS */

#if (defined(THINK_C) || defined(SCOUnix))
#include <time.h>
#else
#include <sys/time.h>
#endif

#include "cdialect.h"

#define byte char

#ifndef cutil_c
extern long myPID;
#endif /* ndef cutil_c */

/*---------------------------------------------------------------------------*/
/* Define boolean data type */
/*---------------------------------------------------------------------------*/

#ifndef boolean
#define boolean	unsigned long
#endif /* ndef boolean */

#ifndef true
#define true    ((boolean)1)
#endif /* ndef true */

#ifndef false
#define false	((boolean)0)
#endif /* ndef false */

/*---------------------------------------------------------------------------*/
/* Math utilities */
/*---------------------------------------------------------------------------*/

#ifndef MAXLONG
#define MAXLONG ((long)0x7FFFFFFF) /* 32 bit longs */
#endif 

#ifndef MINLONG
#define MINLONG ((long)0x80000000) /* 32 bit longs */
#endif

#ifndef MAX
#define MAX(x,y) (((x) > (y)) ? (x) : (y))
#endif /* ndef MAX */
#ifndef MIN
#define MIN(x,y) (((x) < (y)) ? (x) : (y))
#endif /* ndef MIN */
#ifndef ABS
#define ABS(x) (((x) < 0) ? (-(x)) : (x))
#endif /* ndef ABS */

/* XXX any reason why we used to undef instead of ifndef ?
#undef  MAX
#define MAX(x,y) (((x) > (y)) ? (x) : (y))
#undef  MIN
#define MIN(x,y) (((x) < (y)) ? (x) : (y))
#undef  ABS
#define ABS(x) (((x) < 0) ? (-(x)) : (x))
*/

boolean isValidBigNum(char* word);
boolean bigNumEQ(char* v1,char* v2);
boolean bigNumLE(char* v1,char* v2);
boolean bigNumLT(char* v1,char* v2);
boolean bigNumGE(char* v1,char* v2);
boolean bigNumGT(char* v1,char* v2);
long bigNumCmp(void** v1,void** v2);

long    longCmp(void** s1,void** s2);

/*---------------------------------------------------------------------------*/
/* String utlities */
/*---------------------------------------------------------------------------*/

#define STREQ(s1,s2) ((*(s1)==*(s2)) && !strcmp(s1,s2))
#define STRNCMP(s1,s2,n) \
  ((*(s1)==*(s2)) ? strncmp(s1,s2,n) : (*(s1) - *(s2)))

#define safeStrncat(dst,src,maxToAdd,maxTotal)	\
  doSafeStrncat((dst),(src),(maxToAdd),(maxTotal))
#define safeStrncpy(s1,s2,n) doSafeStrncpy((s1),(s2),(n))

#define safeIsAlNum(c) ((c > '`' && c < '{') || \
			(c > '/' && c < ':') || \
			(c > '@' && c < '['))

char* 	doSafeStrncat(char* dst,char* src,size_t maxToAdd,size_t maxTotal);
char* 	doSafeStrncpy(char* s1,char* s2,long n);
long    safeStrlen(char* s);
char* 	safeStrdup(char* s);
char* 	safeStrndup(char* s, long n);
char* 	safeStrAppend(char* s,char* ap);
char*	strtokf(char* s1,long (*isDelimiter)(long c)); 
char*   strtokfIsalnum(char* s1);
boolean substrcmp(char *string1,char *string2);
char    charDowncase(unsigned long ch);
char*   stringDowncase(char* word);
long    alphaCmp(void** s1,void** s2);
long    revAlphaCmp(void** s1,void** s2);
long    longStrCmp(void** s1,void** s2);
long    revLongStrCmp(void** s1,void** s2);
long    longCmp(void** s1,void** s2);

/*---------------------------------------------------------------------------*/
/* Heap utilities */
/*---------------------------------------------------------------------------*/

#define safeMalloc(size)	doSafeMalloc(size)
#define safeRealloc(ptr,size)	doSafeRealloc((ptr),(size))
#define safeFree(ptr)		{ doSafeFree((char*)ptr); (ptr) = NULL; }

void*	doSafeMalloc(size_t size);
void*	doSafeRealloc(void* ptr,size_t size);
void	doSafeFree(void* ptr);

#ifndef cutil_c
extern long md; /* if > 0, and MALLOC_DEBUG, print heap activity */
#endif  /* def cutil_c */

/*---------------------------------------------------------------------------*/
/* Memory Block Utilities */
/*---------------------------------------------------------------------------*/

typedef struct any 
{ unsigned long	size; 
  char*		bytes;
} any;
	
any* makeAny(unsigned long size,char* data);
void freeAny(any* a);
any* duplicateAny(any* a);
any* stringToAny(char* s);
char* anyToString(any* a);
any* longToAny(long Num);
long anyToLong(any* a);
boolean anycmp(any* a1, any* a2);


/*---------------------------------------------------------------------------*/
/* Miscelaneous Utilities */
/*---------------------------------------------------------------------------*/

void initMyPID(void);
void pauseForUser(char* message);
char* nextArg(long* argc,char*** argv);
char* peekArg(long* argc,char*** argv);
char* readNextArg(long* argc,char*** argv,boolean* useFile,FILE* file);
void beFriendly(void);

#ifndef NULL
#define NULL	(0L)
#endif /* ndef NULL */

/*---------------------------------------------------------------------------*/
/* Printing utilities */
/*---------------------------------------------------------------------------*/

#define NL() printf("\n")

long cprintf(boolean print,char* format,...);
void printHexByte(unsigned char h);
char* commafy(long val,char* buf);
void printTime(FILE* f,time_t t);
char* printableTime(time_t t);
char* printableTimeNow(void);

/*---------------------------------------------------------------------------*/
/* Timer utilities */
/*---------------------------------------------------------------------------*/

boolean startTimer(long timerNum);
boolean stopTimer(long timerNum);
long timerVal(long timerNum);

/*---------------------------------------------------------------------------*/


#endif /* ndef cutil_h */


