/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

#include <time.h>

#include "date.h"
#include "futil.h"

/*---------------------------------------------------------------------------*/

void
printDate(FILE* f,date* d)
{
  if (f == NULL)
    f = stdout;

  fprintf(f,"y %ld m %ld d %ld h %ld m %ld s %ld\n",
	  d->year,d->month,d->day,d->hour,d->minute,d->second);
}

/*---------------------------------------------------------------------------*/

void   
dateFromTime(date* d,time_t secSince1970)
{
  struct tm* theTM;

  
  theTM = localtime(&secSince1970); 

  
  d->year = theTM->tm_year + 1900;
  d->month = theTM->tm_mon + 1;
  d->day = theTM->tm_mday;
  d->hour = theTM->tm_hour;
  d->minute = theTM->tm_min;
  d->second = theTM->tm_sec;
}

/*---------------------------------------------------------------------------*/

boolean
timeFromDate(date* d,time_t* secSince1970)
{
  struct tm theTM;

  if (d->year < 1970)
    return(false);

  theTM.tm_year = d->year - 1900;
  theTM.tm_mon = d->month - 1;
  theTM.tm_mday = d->day;
  theTM.tm_hour = d->hour - 1;
  theTM.tm_min = d->minute;
  theTM.tm_sec = d->second;

  /* null these out */
  theTM.tm_wday = 0;
  theTM.tm_yday = 0;
  theTM.tm_isdst = 0;

#ifndef __svr4__
  /* these are obsolete on svr4 */
  theTM.tm_zone = NULL;
  theTM.tm_gmtoff = 0;
#endif

  *secSince1970 = mktime(&theTM);
  
  return(true);
}

/*---------------------------------------------------------------------------*/

char*   
dateToYYYYMMDD(date* d)
{
  char* s;

  s = (char*)safeMalloc(sizeof(char)*(YYYYMMDD_LEN+1));
  sprintf(s,"%04ld%02ld%02ld",d->year,d->month,d->day);

  return(s);
}

/*---------------------------------------------------------------------------*/

date*
YYYYMMDDToDate(char* s)
{
  date* d = NULL;

  if (s != NULL)
    { 
      d = (date*)safeMalloc(sizeof(date));
      sscanf(s,"%04ld%02ld%02ld",&d->year,&d->month,&d->day);
    }

  return(d);
}

/*---------------------------------------------------------------------------*/

char*   
dateToYYYYMMDDHHMMSS(date* d)
{
  char* s;

  s = (char*)safeMalloc(sizeof(char)*(YYYYMMDDHHMMSS_LEN+1));
  sprintf(s,"%04ld%02ld%02ld%02ld%02ld%02ld",
	  d->year,d->month,d->day,d->hour,d->minute,d->second);

  return(s);
}

/*---------------------------------------------------------------------------*/

date*
YYYYMMDDHHMMSSToDate(char* s)
{
  date* d = NULL;

  if (s != NULL)
    { 
      d = (date*)safeMalloc(sizeof(date));
      sscanf(s,"%04ld%02ld%02ld%02ld%02ld%02ld",
	     &d->year,&d->month,&d->day,&d->hour,&d->minute,&d->second);
    }

  return(d);
}

/*---------------------------------------------------------------------------*/



#define yearMask            ((long)0x01FFF)
#define monthBits           (4L)
#define monthMask           ((long)0x0F)
#define dayBits             (5L)
#define dayMask             ((long)0x1F)
#define hourBits            (6L)
#define hourMask            ((long)0x3F)
#define minuteBitsPart1     (4L)
#define minuteMaskPart1     ((long)0x0F)
#define minuteBitsPart2     (2L)
#define minuteMaskPart2     ((long)0x03)
#define secondBits          (6L)
#define secondMask          ((long)0x3F)

/*---------------------------------------------------------------------------*/

void
writeDate(FILE* file,date* d)
{
  unsigned long first4Bytes;
  unsigned char lastByte;

  first4Bytes = d->year & yearMask;

  first4Bytes = first4Bytes << monthBits;
  first4Bytes += d->month & monthMask;

  first4Bytes = first4Bytes << dayBits;
  first4Bytes += d->day & dayMask;

  first4Bytes = first4Bytes << hourBits;
  first4Bytes += d->hour & hourMask;

  first4Bytes = first4Bytes << minuteBitsPart1;
  first4Bytes += (d->minute >> minuteBitsPart2) & minuteMaskPart1;

  lastByte = d->minute & minuteMaskPart2;

  lastByte = lastByte << secondBits;
  lastByte += d->second & secondMask;

#ifdef __alpha /* just want 32 bits */
  { unsigned int f = first4Bytes & 0xFFFFFFFF; 
    safeFWrite(&f,sizeof(unsigned int),1,file);
  }
#else
  safeFWrite(&first4Bytes,sizeof(unsigned long),1,file);
#endif

  safePutc(lastByte,file);
}

/*---------------------------------------------------------------------------*/

void
readDate(FILE* file,date* d)
{
  unsigned long first4Bytes;
  unsigned char lastByte;

  if (d == NULL) 
   { safeFSeek(file,5L,SEEK_CUR);
     return;
   }

#ifdef __alpha /* just want 32 bits */
  { unsigned int f; 
    fread(&f,sizeof(unsigned int),1,file);
    first4Bytes = f;
  }
#else
  fread(&first4Bytes,sizeof(unsigned long),1,file);
#endif

  lastByte = getc(file);

  d->second = lastByte & secondMask;

  d->minute = first4Bytes & minuteMaskPart1;
  d->minute = d->minute << minuteBitsPart2;
  lastByte = lastByte >> secondBits;
  d->minute += lastByte;
  
  first4Bytes = first4Bytes >> minuteBitsPart1;
  d->hour = first4Bytes & hourMask;

  first4Bytes = first4Bytes >> hourBits;
  d->day = first4Bytes & dayMask;

  first4Bytes = first4Bytes >> dayBits;
  d->month = first4Bytes & monthMask;
  
  first4Bytes = first4Bytes >> monthBits;
  d->year = first4Bytes & yearMask;
}

/*---------------------------------------------------------------------------*/

date*   
latestDate(date* d1,date* d2)
/* returns d1 if *d1 is more recent, or equal to *d2.  This lets us use
   latestDate(d1,d2) == d1 to see if d1 >= d2
 */
{
  if (eqDate(d1,d2) ||
      (d1->year > d2->year) ||
      (d1->year == d2->year && d1->month > d2->month) ||
      (d1->year == d2->year && d1->month == d2->month && d1->day > d2->day) ||
      (d1->year == d2->year && d1->month == d2->month && d1->day == d2->day &&
       d1->hour > d2->hour) ||
      (d1->year == d2->year && d1->month == d2->month && d1->day == d2->day &&
       d1->hour == d2->hour && d1->minute > d2->minute) ||
      (d1->year == d2->year && d1->month == d2->month && d1->day == d2->day &&
       d1->hour == d2->hour && d1->minute == d2->minute && 
       d1->second > d2->second))
    return(d1);
  else
    return(d2);
}

/*---------------------------------------------------------------------------*/

boolean
eqDate(date* d1,date* d2)
/* returns true if d1 and d2 are equal */
{
  if (d1->year == d2->year && 
      d1->month == d2->month && 
      d1->day == d2->day &&
      d1->hour == d2->hour &&
      d1->minute == d2->minute && 
      d1->second == d2->second)
    return(true);
  else
    return(false);
}

/*---------------------------------------------------------------------------*/

void    
dateNow(date* d)
{
  time_t timeInSecSince1970;

  if (d == NULL)
    return;

  timeInSecSince1970 = time(NULL);
  dateFromTime(d,timeInSecSince1970);
}

/*---------------------------------------------------------------------------*/

date* 
copyDate(date* d)
{
  date* newDate;

  if (d == NULL)
    return(NULL);

  newDate = (date*)safeMalloc(sizeof(date));
  memcpy(newDate,d,sizeof(date));

  return(newDate);
}

/*---------------------------------------------------------------------------*/

boolean
isValidYYYYMMDDHHMMSSDate(char* word)
/* see if the word has any hope of being a date */
{
  long i;
  for (i = 0; i < YYYYMMDDHHMMSS_LEN; i++)
   { if (!isdigit(word[i]))
       return(false);
   }
  return(true);
}

/*---------------------------------------------------------------------------*/

/*#define testing*/

#ifdef testing

void 
conversionTest(void)
{
  date d1,d2;
  time_t t;

  dateNow(&d1);
  timeFromDate(&d1,&t);
  dateFromTime(&d2,t);
  
  printDate(NULL,&d1);
  printTime(NULL,t); printf("\n");
  printDate(NULL,&d2);
}

void
readWriteTest(void)
{
  date d;
  time_t t;
  FILE* foo;
  long i;

  foo = safeFOpen("foo","wb");
  for (i = 0; i < 5; i++)
   { long sleepTime;
     t = time(NULL);
     dateFromTime(&d,t);
     writeDate(foo,&d);
     printf("t %ld y %ld m %ld d %ld h %ld m %ld s %ld\n",t,d.year,d.month,
	    d.day,d.hour,d.minute,d.second);
     sleepTime = rand() % 20;
     printf("sleeping for %ld seconds\n",sleepTime);
     sleep(sleepTime);
   }
  safeFClose(foo);

  printf("\n");

  foo = safeFOpen("foo","rb");
  for (i = 0; i < 5; i++)
   { readDate(foo,&d);
     timeFromDate(&d,&t);
     printf("t %ld y %ld m %ld d %ld h %ld m %ld s %ld\n",t,d.year,d.month,
	    d.day,d.hour,d.minute,d.second);
   }
  safeFClose(foo);

  printf("\n");
}

main()
{
  conversionTest();
}

#endif 

/*---------------------------------------------------------------------------*/


