/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*-------------------------------------------------------------------------- */
/*-------------------------------------------------------------------------- */

#ifdef THINK_C

#include <pascal.h>		
#if THINK_C == 1
#include <FileMgr.h>
#endif
#if THINK_C == 5
#include <Files.h>
#endif

#else

#include <sys/types.h> 
#include <sys/stat.h> 
#include <sys/param.h> 
#include <sys/file.h>
#include <sys/fcntl.h>
#include <sys/errno.h>
#include <errno.h> /* for solaris ZZZ */

#if (defined(__alpha) || defined (bsdi))
#include <sys/mount.h>
#else
#include <sys/vfs.h>
#endif

#if (defined(__svr4__) || defined(SCOUnix) || defined(hpux))
#include <unistd.h>
#endif

#endif 

#include <string.h>
#include "futil.h"
#include "waislog.h"
#include "panic.h"

#ifdef NeXT 
/* sys/stat.h does not define values for execute permission on 
   group or other 
 */
#define S_IXOTH 0000001 /* other */
#define S_IXGRP 0000010 /* group */
#define S_IXUSR 0000100 /* user */
#endif /* def NeXT */

#ifdef __svr4__
/* solaris is moving away from flock */
#define	LOCK_SH		1	/* shared lock */
#define	LOCK_EX		2	/* exclusive lock */
#define	LOCK_NB		4	/* don't block when locking */
#define	LOCK_UN		8	/* unlock */
#endif /* def __svr4__ */


/*---------------------------------------------------------------------------*/

static long numFilesCurrentlyOpen = 0;
static long maxNumFilesOpenAtOneTime = 0;

FILE*
doSafeFOpen(char* filename,char* mode)
/* 
   note that we follow symbolic links.  If a link is opened in write mode
   (which would normally destroy the link), we instead open in r+ mode 
   which keeps the link, and we sets the size of the link to 0.

   we assume the mode string is always of the form:

   "r|w|a [+] [b]"
 */
{
  FILE* file = NULL;
  char realMode[10];
  boolean openingLinkForWrite = false;

  strcpy(realMode,mode);
  
  if (linkP(filename))
   { if (*mode == 'w')
      { openingLinkForWrite = true;
	strcpy(realMode,"r+");
	if (strchr(mode,'b'))
	  strcat(realMode,"b");
      }
   }

  file = fopen(filename,realMode);

  if (file != NULL)
    { numFilesCurrentlyOpen++;
      if (numFilesCurrentlyOpen > maxNumFilesOpenAtOneTime)
	maxNumFilesOpenAtOneTime = numFilesCurrentlyOpen;
      if (openingLinkForWrite)
	truncate(filename,0);
    }

/*printf("fopen %lx %s %s\n",file,filename,mode);*/

/*printf("fopen reports numFilesCurrentlyOpen %ld\n",numFilesCurrentlyOpen);*/

  return(file);
}

/*---------------------------------------------------------------------------*/

long
doSafeFClose(FILE* file)
{

/*printf("fclose %lx\n",file);*/

  if (file != NULL)
    { numFilesCurrentlyOpen--;
      return(fclose(file));
    }
  else
    return(0L);
}

/*---------------------------------------------------------------------------*/

long 
doSafeFSeek(FILE* file,long offset,long fromWhere)
{
  long result;
  
  if (file == NULL)
    return(-1);
  
/*cprintf(dbg,"seek %ld %ld %ld\n",fileno(file),offset,fromWhere);*/

  if (offset == 0 && fromWhere == SEEK_CUR) /* this is a nop */
    return(0);

  result = fseek(file,offset,fromWhere);

/*  if (result != 0) 
    cprintf(dbg,"seek failed:  file %ld offset %ld, fromWhere %d\n",
	    fileno(file),offset,fromWhere);
*/

  return(result);
}

/*---------------------------------------------------------------------------*/

long 
doSafeFTell(FILE* file)
{
  long result;
  
  if (file == NULL)
    return(0);
    
  result = ftell(file);
  
  if (result == EOF)
    panic("ftell failed\n");

  return(result);
}

/*---------------------------------------------------------------------------*/

long 
doSafeRename(char* path1,char* path2)
{

  /* if either of the paths are through links we want to handle them specialy
     so that we don't lose the link and can go across file systems
     NOTE - the special handling is to copy the file which may be very
     slow and take up space.  Use with caution.
   */
  if (linkP(path1) || linkP(path2))
   { FILE* from = safeFOpen(path1,"rb");
     FILE* to = safeFOpen(path2,"wb");
     long i;
     long ret;

     if (strcmp(path1,path2) == 0) /* wish there was a better equiv test */
       ret = 0; /* they are the same file */
     else if (to == NULL || from == NULL)
	ret = -1; /* error opening the file */
     else
      { while ((i = getc(from)) != EOF)
	  safePutc(i,to);
	truncate(path1,0);
	if (!linkP(path1))
	  remove(path1);
	ret = 0;
      }

     safeFClose(to);
     safeFClose(from);

     return(ret);
   }
  else /* let the OS do it */
   {
#ifdef THINK_C
     remove(path2);
#endif  

     return(rename(path1,path2));
   }
}

/*---------------------------------------------------------------------------*/

boolean
retryTillDiskSpaceAvailable(FILE* fileNeedingSpace,long spaceNeeded)
/* loop until the disk frees up, writing log messages as we go */
{
  boolean waitForSpace = true; /* XXX should be setable */

  if (waitForSpace) 
   {
     /* XXX change name of executable to */
     /* XXX set up to handle a shutdown signal? */
     
     while (freeBytesOnFileSystem(fileNeedingSpace) < spaceNeeded)
      { waislog(WLOG_HIGH,WLOG_INFO,"waiting for %ld bytes on disk...",
		spaceNeeded); 
	/* XXX this assumes the disk with the log is not full, and someone
	   is reading the long... sigh */
	/* XXX print name of disk requiring space */
	sleep(60);
      }
     
     return(true);
   }
  else /* don't wait for the space */
    return(false);
}

/*---------------------------------------------------------------------------*/

long    
doSafeFWrite(void* data,long size,long num,FILE* file)
{
  long res;

  if (file == NULL)
    panic("safeFWrite:  writing to NULL file\n");

  res = fwrite(data,size,num,file);

  if (res == 0)
   { /* error */
     if (freeBytesOnFileSystem(file) == 0)
      { if (retryTillDiskSpaceAvailable(file,num*size))
	  return(doSafeFWrite(data,size,num,file));
        else
	  panic("out of disk space");
      }
     else
       panic("write error %ld",errno);
   }
  
  return(res);
}

/*---------------------------------------------------------------------------*/

long    
doSafePutc(char data,FILE* file)
{
  long res;

  if (file == NULL)
    panic("safePutc:  writing to NULL file\n");

  res = putc(data,file);

  if (res == EOF)
   { /* error */
     if (freeBytesOnFileSystem(file) == 0)
      { if (retryTillDiskSpaceAvailable(file,100))
	  return(doSafePutc(data,file));
	else
	  panic("out of disk space");
      }
     else
       panic("write error %ld",errno);
   }

  return(res);
}

/*----------------------------------------------------------------------*/

long
doSafeFPrintfInternal(FILE* file,char* format,va_list ap)
{
  long res;

  if (file == NULL)
    panic("safeFPrintf:  writing to NULL file\n");
  
  res = vfprintf(file,format,ap); 
  if (res == EOF)
   { /* error */
     if (freeBytesOnFileSystem(file) == 0)
      { if (retryTillDiskSpaceAvailable(file,5000))
	  return(doSafeFPrintfInternal(file,format,ap));
	else
	  panic("out of disk space");
      }
     else
       panic("write error %ld",errno);
   }

  va_end(ap); 
}

/*----------------------------------------------------------------------*/

long
safeFPrintf(FILE* file,char* format,...)
{
  va_list ap; 
  long res;

  va_start(ap,format);

  return(doSafeFPrintfInternal(file,format,ap));

  va_end(ap); 
}

/*---------------------------------------------------------------------------*/

char*   
doSafeFGets(char* data,long size,FILE* file)
/*
  This routine turns data read from fgets into a proper c string and 
  removes the final \n.

  Warning - the output of standard fgets is not a cstring!  It's a buffer 
  containing whatever was in the file, terminated by either

   - a line feed (\n) or 
   - at the very end of the buffer - a '\0'

  In other words, it can contain nulls.  Since C uses the null character as a 
  string terminator, we need to convert them to something C does not use
  as a terminator - space will do.  Howerver, if a newline cannot be found, 
  then s[len - 1] is a null. Therefore, do NOT convert s[len - 1] to a ' '!

  Also, we need to look out for data which is not broken into lines.  Fgets
  will read up to the end of the data, and not terminate it with either
  a \n, or a null.  We solve this by initing the entire line to \n before
  reading.  (this was the notoriuos stanford bug).

  XXX maybe we should consider simply implementing our own fgets like function
  rather than working around the existing one.
*/
{
  char* res;
  long i;

  /* set the line to \n's, which is the fgets terminator, thus any data
     not written by fgets will automatically show up as a terminator and
     be turned into a C terminator in the final pass of this routine */
  memset(data,'\n',size); 

  /* read the line */
  res = fgets(data,size,file);

  /* convert any \n's to C terminators ('\0'), convert any C terminators
     to spaces.  Do not convert the very last character in the buffer */
  if (res != NULL)
   { for (i = 0; i < size - 1; i++)
      { if (data[i] == '\n') 
	 { /* \n is the fgets terminator, replace with the C terminator
	      and exit */
	   data[i] = '\0';
	   break;
	 }
        else if (data[i] == '\0') 
	 { /* \0 is not a terminator from fgets point of view, but it is for
	      C.  Convert it into something that C won't barf on */
	   data[i] = ' ';
	 }
      }
   }
  return(res);
}

/*---------------------------------------------------------------------------*/

long 
writeBytes(long value,long size,FILE* stream)
{
  long i;
  long answer;

  if ((size < sizeof(long)) && ((value >> (size * 8)) != 0))
    panic("In a call to writeBytes, the value %ld can not be represented in %ld bytes",value,size);

  for (i = size - 1; i >= 0; i--)
    answer = safePutc((value >> (i * 8)) & 0xFF,stream);

  if (ferror(stream) != 0) 
    panic("Write failed");

  return(answer);
}

/*---------------------------------------------------------------------------*/

long 
readBytes(long n,FILE* stream)
{
  long answer = 0;
  unsigned long ch;
  long i;

  for (i = 0; i < n; i++)
   { ch = getc(stream);
     if (ch == EOF)
      return(EOF);
     answer = (answer << 8) + (unsigned char)ch;
  }
  return(answer);
}

/*---------------------------------------------------------------------------*/

long 
readBytesFromMemory(long n,unsigned char* block)
{
  long answer = 0;
  unsigned char ch;
  long i;

  for (i = 0; i < n; i++)
   { ch = *(block++);
     answer = (answer << 8) + ch;
   }
  return(answer);
}

/*---------------------------------------------------------------------------*/

static char*
cleanPath(char* filename)
{

#ifdef THINK_C
  return(filename);
#endif /* def THINK_C */

  char* beginningPtr = strstr(filename,"/../");
  if (beginningPtr != NULL)
   { char *ptr;
     for (ptr = beginningPtr - 1; ptr >= filename; ptr--)
      { if (*ptr == '/')
	 { strcpy(ptr,beginningPtr + strlen("/../") - 1);
	   cleanPath(filename);	
	   break;
	 }	
      }
   }
  
  beginningPtr = strstr(filename,"/./");
  if (beginningPtr != NULL)
   { strcpy(beginningPtr,beginningPtr + strlen("/./") - 1);
     cleanPath(filename);	
   }

  return(filename);
}

/*---------------------------------------------------------------------------*/
    
char*
truename(char* filename,char* fullPath)
{
  if (filename == NULL || fullPath == NULL)
    return(NULL);

#ifdef THINK_C

  strcpy(fullPath,filename);
  return(fullPath); 

#else

  if (strcmp(filename,"<stdin>") == 0)
    strcpy(fullPath,filename);
  else if (filename[0] == '/')
   { strcpy(fullPath,filename);
     cleanPath(fullPath);
     return(fullPath);
   }
  else
   { getwd(fullPath);
     safeStrncat(fullPath,"/",MAX_FILENAME_LEN,MAX_FILENAME_LEN);
     safeStrncat(fullPath,filename,MAX_FILENAME_LEN,MAX_FILENAME_LEN);
     cleanPath(fullPath);
     return(fullPath);
   }
#endif 
}

/*---------------------------------------------------------------------------*/

char*
pathnameName(char* pathname)
{

#ifdef THINK_C
  char *answer = strrchr(pathname,':');
#else
  char *answer = strrchr(pathname,'/');
#endif	

  if (answer == NULL)
    return(pathname);
  return(answer + 1);
}

/*---------------------------------------------------------------------------*/

char*
pathnameDirectory(char* pathname,char* destination)
{
#ifdef THINK_C
n  char *dirptr = strrchr(pathname,':');
#else
  char *dirptr = strrchr(pathname,'/');
#endif 

  if (dirptr == NULL)
   {
#ifdef THINK_C
     strncpy(destination,pathname,MAX_FILENAME_LEN);
#else
     strncpy(destination,"./",MAX_FILENAME_LEN);
#endif 
   }
  else
    { strncpy(destination,pathname,MAX_FILENAME_LEN);
      destination[dirptr - pathname + 1] = '\0';
    }

  return(destination);
}
  
/*---------------------------------------------------------------------------*/

char*
mergePathnames(char* pathname,char* directory)
{
  static char answer[MAX_FILENAME_LEN + 1];

  if ((pathname[0] == '/') || (NULL == directory) || directory[0] == '\0')
    return(pathname);
  else
   { answer[0] = '\0';
     strncat(answer,directory,MAX_FILENAME_LEN);

#ifdef THINK_C
    if (directory[strlen(directory) - 1] != ':')
      strncat(answer,":",MAX_FILENAME_LEN);
#else
    if (directory[strlen(directory) - 1] != '/')
      strncat(answer,"/",MAX_FILENAME_LEN);
#endif
    strncat(answer,pathname,MAX_FILENAME_LEN);
  }
  
  return(answer);
}

/*---------------------------------------------------------------------------*/

boolean 
directoryP(char* f)
/* XXX strangely, this fails if f = "".  It says that f is a directory */
{
#ifdef THINK_C
  return(false); /* should check if it is a folder */
#else
  struct stat stbuf;
  if (f == NULL || strcmp(f,"") == 0)
    return(false);
  else if (stat(f,&stbuf) != 0)
    return(false);
  if ((stbuf.st_mode & S_IFMT) == S_IFDIR)
    return(true);
  return(false);
#endif
}

/*---------------------------------------------------------------------------*/

boolean 
fileP(char* f)
{
#ifdef THINK_C 
  return(probeFile(f));
#else
  struct stat stbuf;
  if (f == NULL || strcmp(f,"") == 0)
    return(false);
  if (stat(f,&stbuf) == -1)
    return(false);
  if (!((stbuf.st_mode & S_IFMT) == S_IFDIR))
    return(true);
  return(false);
#endif
}

/*---------------------------------------------------------------------------*/

boolean
linkP(char* f)
/* return true if filename contains a symbolic link */
{

#ifdef THINK_C
  return(false); /* should check for aliases */
#else
  struct stat stbuf;
  if (f == NULL || strcmp(f,"") == 0)
    return(false);
  if (lstat(f,&stbuf) != 0)
    return(false);
  if ((stbuf.st_mode & S_IFLNK) == S_IFLNK)
    return(true);
  return(false);
#endif
}

/*---------------------------------------------------------------------------*/

boolean
executableP(char* f)
/* return true if file is an executable program runnable by the current user */
{
#ifdef THINK_C
  return(false); /* should check for mac type APPL */
#else
  struct stat stbuf;
  if (f == NULL || strcmp(f,"") == 0)
    return(false);
  if (stat(f,&stbuf) != 0)
    return(false);
  if (stbuf.st_mode & S_IXOTH) /* executable by other */
    return(true);
  if (stbuf.st_gid == getgid() && stbuf.st_mode & S_IXGRP) /* group */
    return(true);
  if (stbuf.st_uid == getuid() && stbuf.st_mode & S_IXUSR) /* user */
    return(true);
  return(false);
#endif
}

/*---------------------------------------------------------------------------*/

long    
getFilePermission(char* f)
{
  if (f == NULL || strcmp(f,"") == 0)
    return(NO_PERMISSION);

#ifdef THINK_C
  /* I'm sure there is a better way, since we should have to deal with
     appleshare etc */
  return(READ_PERMISSION | WRITE_PERMISSION);
#else
  { struct stat stbuf;
    if (stat(f,&stbuf) != 0)
      return(NO_PERMISSION);
    else 
     { long perm;
/*
       printf("  p %lX w %lX g %lX u %lX\n",stbuf.st_mode,stbuf.st_mode & 07,
	      (stbuf.st_mode >> 3) & 07,(stbuf.st_mode >> 6) & 07);
       printf("  g %ld sg %ld u %ld su %ld\n",getegid(),stbuf.st_gid,
	      geteuid(),stbuf.st_uid);
*/
       perm = stbuf.st_mode & 0x0F; /* world permissions */
       if (getegid() == stbuf.st_gid)
	 perm |= ((stbuf.st_mode >> 3) & 07); /* group perms */
       if (geteuid() == stbuf.st_uid)
	 perm |= ((stbuf.st_mode >> 6) & 07); /* user perms */
       return(perm);
     }
  }
#endif
}

/*---------------------------------------------------------------------------*/

boolean 
probeFile(char* f)
{
  if (f == NULL || strcmp(f,"") == 0)
    return(false);

#ifdef THINK_C  

 { FILE* f;
   f = safeFOpen(f,"r");
   if (f == NULL)
     return(false);
   else
    { safeFClose(f);
      return(true);
    }
 }

#else

  if (access(f,R_OK) == 0)
    return(true);
  else
    return(false);

#endif 
}

/*---------------------------------------------------------------------------*/

boolean 
probeFilePossiblyCompressed(char* f)
{
  if (f == NULL || strcmp(f,"") == 0)
    return(false);
  else if (!probeFile(f))
    { char buffer[MAX_FILENAME_LEN * 2];
      strcpy(buffer,f);
      strcat(buffer,".Z");
      return(probeFile(buffer));
    }
  else
    return(true);
}

/*---------------------------------------------------------------------------*/

boolean 
touchFile(char* f)
{
  FILE *stream = NULL;
  if (f == NULL)
    return(false);
  stream = safeFOpen(f,"a");
  if (stream == NULL)
    return(false);
  else
    { safeFClose(stream);
      return(true);
    }
}

/*---------------------------------------------------------------------------*/

long 
fileLength(FILE* stream)
{
  long position;
  long end = -1;
  if (stream == NULL)
    return(-1);
  position = ftell(stream);
  safeFSeek(stream,0L,SEEK_END);
  end = ftell(stream);	
  safeFSeek(stream,position,SEEK_SET);
  return(end);
}

/*---------------------------------------------------------------------------*/

#if (!defined(__svr4__) && !defined(bsdi))

char* 
strerror(long error) 
/* this should take a unix error code and return the string describing it,
   just as perror prints the code out.  Unfortunately only some unicies
   don't support this nice feature 
 */
{ 
  static char msg[100];
  sprintf(msg,"errno %ld",error);
  return(msg); 
} 

#endif 

/*---------------------------------------------------------------------------*/

void
getFileDates(char* f,date* createDate,date* modDate)
{ 
#ifdef THINK_C
  /* XXX ought to be a way */
  if (createDate != NULL)
    *createDate = 0;
  if (modDate != NULL)
    *modDate = 0;
#else
  struct stat buf;
  date theDate;

  if (stat(f,&buf) != 0)
    panic("could not stat %s: %s",f,strerror(errno));

  dateFromTime(&theDate,buf.st_mtime);

  if (modDate != NULL)
    *modDate = theDate;
  if (createDate != NULL)
    *createDate = theDate; /* no separate creation date maintained in 
			      unix! */
#endif 
}

/*---------------------------------------------------------------------------*/

time_t
getOpenFileModTime(FILE* f)
{ 

#ifdef THINK_C
  /* XXX ought to be a way */
  return(0);
#else

  struct stat buf;

  if (f == NULL)
    return(0);

  if (fstat(fileno(f),&buf) != 0)
    panic("could not stat open file: %s",strerror(errno));

  return(buf.st_mtime);
#endif 
}

/*---------------------------------------------------------------------------*/

void
checkForBlocks(long f,struct flock* lockRequested)
/* this was for testing locks, it is no longer used, but we'll keep it around
   for a while
 */
{
  struct flock lck;
  long res;

  lck = *lockRequested;
  errno = -1;

  res = fcntl(f,F_GETLK,&lck); /* see if there is anything that will block */
  if (res == 0)
   {  if (lck.l_type == F_UNLCK)
	printf("no blockage\n"); /* not locked, we are free to go */
      else if (lck.l_pid == getpid())
       { /* it's my lock, unlock it */
	 printf("will be blocked by lock %ld from pid %ld.  unlocking...\n",
		lck.l_type,lck.l_pid);
	 lck.l_type = F_UNLCK;
	 lck.l_whence = lck.l_start = lck.l_len = 0; /* lock the whole file */
	 errno = -1;
	 res = fcntl(f,F_SETLK,&lck); /* unlock it */       
	 printf("unlock returned %ld errno %ld\n",res,errno);
       }
      else
	printf("waiting on someone elses lock\n");
    }
  else
    printf("couldn't get current lock stat from %ld, res %ld errno %ld\n",
	   f,res,errno);
}

/*---------------------------------------------------------------------------*/

long 
portableFLock(long f,long lockType)
/* fcntl is the low level routine  seemingly used by flock and lockf. 
   The idea here is to make it work like flock, and see if that is 
   supported on the "modern" OS's...

   ok, this seems to work on solaris, and even works over nfs apparently.
   it's also a ton cleaner than the lockf solution

   yay! 

   note - one problem - in order to have a read lock you need to have the
   file open with read permission.  In order to have an exclusive lock
   you need to have the file open with write permission.  sigh.
 */
{
  long res;
  long cmd;
  struct flock lck;

  if (lockType & LOCK_NB) /* non blocking */
   { cmd = F_SETLK;
     lockType = lockType & ~LOCK_NB; /* turn off non blocking bit */
   }
  else /* blocking */
    cmd = F_SETLKW;

  /* lock type is as per flock() */
  if (lockType == LOCK_SH)
    lck.l_type = F_RDLCK;
  if (lockType == LOCK_EX)
    lck.l_type = F_WRLCK;
  if (lockType == LOCK_UN)
    lck.l_type = F_UNLCK;

  lck.l_whence = lck.l_start = lck.l_len = 0; /* lock the whole file */

/*
printf("fcntl setting lock on %ld to %ld\n",f,lck.l_type);

if (lockType != LOCK_UN)
 { checkForBlocks(f,&lck);
   checkForBlocks(f,&lck);
 }

printf("ok, here goes, the file is %ld, the cmd is %ld and the lock type is %ld\n",f,cmd,lck.l_type);
*/

  errno = -1;
  res = fcntl(f,cmd,&lck); /* lock it */

/*
if (res != 0)
 { printf("ug, res = %ld errno = %ld, lets see if there is any blocking\n",
	  res,errno);
   checkForBlocks(f,&lck);
 }
*/
  return(res);
}

/*---------------------------------------------------------------------------*/

boolean
lockFile(FILE* f,boolean isExclusive,long secsToWait)
/* 
   Place a lock on a file f.  If isExclusive is false, it is an advisory
   lock, which is intended to say "we are reading this file".  Otherwise it
   is an exclusive lock, which says "we are modifying this file".  There 
   may be many simultaneous advisory locks but only one exclusive lock.  
   Furthermore, the two kinds of locks are mutually exclusive.
   SecsToWait <= 0 means block (ie. wait indefinitely) while waiting for a 
   lock.  Otherwise the routine returns true if the lock was successful,
   false otherwise.
 */
{
  long operation;

/*printf("lockFile %lx isExclusive %ld secsToWait %ld\n",f,
       isExclusive,secsToWait);*/

  if (f == NULL)
    return(false);

  if (isExclusive)
    operation = LOCK_EX;
  else
    operation = LOCK_SH;    

  if (secsToWait > 0) /* we don't want to wait forever */
   { time_t startTime;
     operation |= LOCK_NB; /* non blocking */
     for (startTime = time(NULL); startTime + secsToWait >= time(NULL); )
      { if (portableFLock(fileno(f),operation) == 0)
	  return(true);

/*
printf("errno %ld t  %ld\n",errno,errno == 0 || errno == EWOULDBLOCK || errno == EAGAIN || errno == EINTR || errno == ENOLCK);
*/

	if (errno == 0 || errno == EWOULDBLOCK || errno == EAGAIN ||
	    errno == EINTR || errno == ENOLCK)
	 { /* should we sleep here? rather than going into a tight loop?
	      originally we didn't but that means that several programs
	      can start waiting for locks and really eat up cpu.  the
	      problem with sleeping is that if you have N processes waiting
	      for the lock it becomes failry random which one will get it
	    */
	   sleep(1L);
	   continue; /* else keep trying until timer runs out */
	 }
	else
	  return(false); /* flock failed */
      }
     return(false);
   } 
  else /* block */
   { if (portableFLock(fileno(f),operation) == 0)
       return(true);
     else
       return(false);
   } 
}

/*---------------------------------------------------------------------------*/

void
unlockFile(FILE* f)
/* undoes the work of lockFile().  see comments there */
{
/*printf("unlockFile %lx\n",f);*/

  if (f == NULL)
    return;

  portableFLock(fileno(f),LOCK_UN);
}

/*---------------------------------------------------------------------------*/

long 
freeBytesOnFileSystem(FILE* f)
/* return the number of bytes on the file system containing f */
{
#ifndef __svr4__
/* ZZZ solaris doesn't know about this */

  struct statfs buf;

  if (fstatfs(fileno(f),&buf) == 0) /* success */
    return(buf.f_bsize * buf.f_bavail);
  else
   { /* should handle error conditions in errno */ 
     return(0);
   }
#else /* ndef __svr4 */
  return(0);
#endif /* else */
}

/*---------------------------------------------------------------------------*/

char*
currentUserName(void)
{
  static char answer[200];
  char hostname[120];
  
#ifdef THINK_C

  strcpy(answer,"MAC"); 

#else

#include <pwd.h>  

  struct passwd *pwent = getpwuid(getuid());
  strncpy(answer,pwent->pw_name,200);
  strncat(answer,"@",200);
  gethostname(hostname,120);
  strncat(answer,hostname,200);

#endif 

  return(answer);
}

/*---------------------------------------------------------------------------*/

boolean 
readStringFromFile(FILE* stream,char* array,long arrayLength)
{
  long ch;
  long count = 0;

  array[0] = '\0';
  while (true)
   { ch = getc(stream);
     if(ch == EOF)
      { array[count] = '\0';
	return(false);
      }
    else if (count == arrayLength)
     { array[count] = '\0';
       return(false);
     }
    else if ('\0' == ch)
     { array[count] = '\0';
       return(true);
     }
    else
      array[count++] = ch;
   }
}

/*---------------------------------------------------------------------------*/

FILE* 
openTempFile(char* dir,char* root,char* fullName,char* mode)
/* given dir and mode, return a file pointer to the file, fullName is the
   full name of the file including dir.  The leaf of the name begins with 
   root. Only makes sense if mode is a write or append mode
 */
{
  FILE* f;

  if (dir == NULL)
    dir = ".";

  if (root == NULL)
    root = "temp";

  if (!directoryP(dir))
    return(NULL);

  while (true)
   { sprintf(fullName,"%s/%s%ld.%ld",dir,root,getpid(),random() % 10000);
     if ((f = safeFOpen(fullName,"r")) == NULL)
      { /* no file there to read, we're ok */
	return(safeFOpen(fullName,mode));
      }
     else
       safeFClose(f);
   }
}

/*---------------------------------------------------------------------------*/

void 
growFile(FILE* file,long length)
{
  long currentLength;
  safeFSeek(file,0L,SEEK_END);
  currentLength = safeFTell(file);
  safeFSeek(file,length - currentLength,SEEK_END);
}

/*---------------------------------------------------------------------------*/

long 
removePreservingLink(char* filename)
/* if the file is a link just truncate it (throw away data but keep file)
   otherwise remove it outright
 */
{
  if (linkP(filename))
    return(truncate(filename,0));
  else
    return(remove(filename));
}

/*---------------------------------------------------------------------------*/

