/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

#include "list.h"

/*---------------------------------------------------------------------------*/

void 
mapcar(list l,void (*function)())
{
  long i;

  if (null(l))
    return;

  for (i = 0; i < length(l); i++)
   { (*function)(nth(l,i));
   }
}

/*---------------------------------------------------------------------------*/

list
collecting(list l,void *item)

{
  if (l == NULL)
    l = (list)safeMalloc(sizeof(listStruct));

  if (l->rlen < length(l) + 1)
   { l->rlen = length(l) + 2;
     l->elems = (void**)safeRealloc(l->elems,l->rlen * sizeof(void*));
   }

  l->elems[length(l)] = item;
  l->len++;

  return(l);
}

/*---------------------------------------------------------------------------*/

void
freeList(list l)
{
  if (l != NULL)
   { safeFree(l->elems);
     safeFree(l);
   }
}

/*---------------------------------------------------------------------------*/

void
sortList(list l,long (*cmp)(void** arg1,void** arg2))
{ 
  if (cmp == NULL)
    cmp = alphaCmp;

  if (length(l) > 1)
    qsort(l->elems,length(l),sizeof(void*),cmp);
}

/*---------------------------------------------------------------------------*/

list 
mergeLists(list l1,list l2,long (*cmpFunc)(void** arg1,void** arg2),
	   void (*freeFunc)(void* arg1))
/* given to sorted lists, merge them into a single sorted list.  The input
   lists are freed.  cmp is a function for comparing elements of the lists.
   freeFunc is a function to free elements of the lists which are redundant 
   (if the same element occurs in both lists, the one in l2 is freed, and the
   one in l1 is put in the result list).

   NOTE - the lists must be sorted!
 */
{
  void *v1,*v2;
  list answer = NULL;
  long cmp;

  v1 = pop(&l1);
  v2 = pop(&l2);

  while (true)
   { 
     if (length(l1) == 0 && length(l2) == 0)
      { freeList(l1);
	freeList(l2);
	return(answer); /* done, return the answer */
      }
     
     cmp = cmpFunc(&v1,&v2);

     if (cmp == 0) /* they are the same, use just one */
      { answer = collecting(answer,v1);
	freeFunc(v2);
	v1 = pop(&l1);
	v2 = pop(&l2);
      }
     else if (cmp < 0) /* use the first */
      { answer = collecting(answer,v1);
	v1 = pop(&l1);
      }
     else /* (cmp > 0) use the second */
      { answer = collecting(answer,v2);
	v2 = pop(&l2);
      }
   }
}

/*---------------------------------------------------------------------------*/

list 
insertNth(list l,long n,void* elem)
{
  if (null(l))
    return(NULL);

  if (n > length(l)) 
    return(NULL);

  if (l->rlen < length(l) + 1)
   { l->rlen = length(l) + 2;
     l->elems = (void**)safeRealloc(l->elems,l->rlen*sizeof(void*));
   }

  memmove(l->elems+n+1,l->elems+n,(length(l)-n)*sizeof(void*));
  l->elems[n] = elem;
  l->len++;

  return(l);
}

/*---------------------------------------------------------------------------*/

list
removeNth(list l,long n)
{
  if (null(l))
    return(NULL);

  if (n >= length(l)) 
    return(NULL);

  memmove(l->elems+n,l->elems+n+1,(length(l)-n-1)*sizeof(void*));
  l->len--;
  
  if (l->rlen > 4 && l->rlen > length(l) * 2)
   { l->rlen = length(l);
     l->elems = (void**)safeRealloc(l->elems,l->rlen * sizeof(void*));
   }

  return(l);
}

/*---------------------------------------------------------------------------*/

list 
removeVal(list l,void* val)
{ 
  long i;

  if (null(l))
    return(NULL);

  for (i = 0; i < length(l); i++)
   { if (nth(l,i) == val)
       break;
   }
	
  if (i < length(l)) /* found it */
    return(removeNth(l,i));
  else /* val not in the list */
    return(l);
}

/*---------------------------------------------------------------------------*/

boolean
lookupInSortedList(list l,void* val,long (*cmpFunc)(void** arg1,void** arg2),
		   long* posInList)

{
  long pos;
  long upperBound = length(l);
  long lowerBound = 0;
  long cmp;
  void* data;

  if (null(l))
   { if (posInList != NULL)
       *posInList = 0;
     return(false);
   }

  if (cmpFunc == NULL)
    cmpFunc = alphaCmp;

  while (upperBound >= lowerBound)
   { 
     pos = lowerBound + ((upperBound - lowerBound) / 2);

     if (pos == length(l))
       cmp = -1;
     else
      { data = nth(l,pos);
	cmp = (*cmpFunc)(&val,&data);
      }

     if (cmp == 0)
      { if (posInList != NULL)
	  *posInList = pos;
	return(true);
      }
     else if (cmp < 0) 
      { if (pos == lowerBound) 
	 { if (posInList != NULL)
	     *posInList = pos;
	   return(false);
	 }
	else
	  upperBound = pos; 
      }
     else 
      { if (pos >= length(l))
	 { if (posInList != NULL)
	     *posInList = pos;
	   return(false);
	 }
	else
	  lowerBound = pos + 1; 
      }
   }

  if (posInList != NULL)
    *posInList = length(l); 
  return(false);
}

/*---------------------------------------------------------------------------*/

void*
pop(list* l)
{ 
  void* ans;

  if (l == NULL || *l == NULL)
    return(NULL);

  ans = first(*l);
  *l = removeNth(*l,0);
  return(ans);
}

/*---------------------------------------------------------------------------*/

void 
push(list* l,void* elem)
{
  if (l == NULL)
    return;

  *l = insertNth(*l,0,elem);

}

/*---------------------------------------------------------------------------*/
