/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/* This package lets us read and write stdio through an nConn.  
   XXX it makes stdin non blocking - should it copy stdin?
*/

#include "stdioConn.h"
#include "sockets.h"
#include "waislog.h"

#include <fcntl.h>

typedef struct stdioConn
{ nConn nc;
  FILE* in;
  FILE* out;
} stdioConn;

/*-------------------------------------------------------------------------- */

nConnStatus 
stdioConnInitialize(nConn* conn)
{
  /* don't need to do anything */
  return(nConnStatusOK);
}

/*-------------------------------------------------------------------------- */

nConnStatus 
stdioConnRead(nConn* conn,unsigned char* val)
/* asynchronous read */
{
  stdioConn* c = (stdioConn*)conn;
  long nextChar;
  nextChar = getc(c->in);
  if (nextChar == EOF)
   { if (feof(c->in))
       return(nConnStatusReadEOF);
     else if (ferror(c->in))
       return(nConnStatusReadFailed);
     else
       return(nConnStatusReadNoData);
   }
  else
   { if (val != NULL)
       *val = nextChar & 0xFF;
     return(nConnStatusOK);
   }
}

/*-------------------------------------------------------------------------- */

nConnStatus 
stdioConnPeek(nConn* conn,unsigned char* val)
/* asynchronous read */
{
  stdioConn* c = (stdioConn*)conn;
  long nextChar;
  nextChar = getc(c->in);
  if (nextChar == EOF)
   { if (feof(c->in))
       return(nConnStatusReadEOF);
     else if (ferror(c->in))
       return(nConnStatusReadFailed);
     else
       return(nConnStatusReadNoData);
   }
  else
   { if (val != NULL)
      { *val = nextChar & 0xFF;
	if (ungetc(nextChar,c->in) == EOF)
	  return(nConnStatusReadEOF);
      }
     return(nConnStatusOK);
   }
}

/*-------------------------------------------------------------------------- */

nConnStatus 
stdioConnWrite(nConn* conn,unsigned char val)
{
  long res;
  stdioConn* c = (stdioConn*)conn;

#ifdef __alpha /* this is needed for osf1/v2, and maybe for hpux 8.0 as well */
  { char buf[1];
    buf[0] = val;
    res = write(fileno(c->out),buf,1L);
  }
#else
  res = putc(val,c->out);
#endif

  if (res == EOF)
    return(nConnStatusWriteFailed);
  else
    return(nConnStatusOK);
}

/*-------------------------------------------------------------------------- */

nConnStatus 
stdioConnFlush(nConn* conn)
{
  stdioConn* c = (stdioConn*)conn;
  fflush(c->out);
  return(nConnStatusOK);
}

/*-------------------------------------------------------------------------- */

nConnStatus 
stdioConnFree(nConn* conn)
{
  stdioConn* c = (stdioConn*)conn;

  /* nothing special to do, since we don't want to close stdin/stdout */
  safeFree(conn->name);
  safeFree(c);
  return(nConnStatusOK);
}

/*-------------------------------------------------------------------------- */

nConnStatus
stdioConnGetId(nConn* conn,char** netId,char** name)
/* fill in the netId and name with information about the entity on the
   other side of the connection.  In this case the ip number and host name.
   We malloc the fields.
 */
{
  stdioConn* c = (stdioConn*)conn;
  struct sockaddr_in source;
  long sourcelen;

 if (netId != NULL)
   *netId = NULL;
  if (name != NULL)
    *name = NULL;

  sourcelen = sizeof(struct sockaddr_in);

  /* get information on the machine on the other end of the connection
     (our peer).  note - getpeername returns 0 if successful
   */
  if (!getpeername(fileno(c->in),(struct sockaddr*)&source,&sourcelen))
   { struct hostent* peer = NULL;

     /* we have information, find the peer's hostname */
     peer = gethostbyaddr((char*)&source.sin_addr,4,AF_INET);
     if (peer != NULL && name != NULL)
       *name = safeStrdup(peer->h_name);
     
     /* and internet address */
     if (netId != NULL)
       *netId = safeStrdup(inet_ntoa(source.sin_addr));

     /* if there was no hostname default to the address */
     if (name != NULL && *name == NULL)
       *name = safeStrdup(*netId);

     return(nConnStatusOK);
   }
  else
   { return(nConnStatusNoIDAvailable);
   }
}

/*-------------------------------------------------------------------------- */

nConnStatus 
stdioConnOpen(nConn** conn,long numArgs,va_list ap)
/* make a stdio connection to the specified server and port.  The connection
   does not block on reads
 */
{
  FILE* in;
  FILE* out;
  stdioConn* sc;
  nConn* nc;

  if (numArgs != 2)
    return(nConnStatusBadArguments);

  in = (FILE*)va_arg(ap,FILE*);
  out = (FILE*)va_arg(ap,FILE*);

  sc = (stdioConn*)safeMalloc(sizeof(stdioConn));
  nc = (nConn*)sc;
  sc->in = in;
  sc->out = out;

  if (sc->in != NULL)
   { *conn = (nConn*)sc;
     return(nConnStatusOK);
   }
  else
   { /* XXX handle error */
     return(nConnStatusOpenFailed);
   }
}

/*-------------------------------------------------------------------------- */

nConnStatus 
stdioConnWaitForConnection(nConn** conn,long numArgs,va_list ap)
/* this does not make sense for stdio */
{ 
  return(nConnStatusUndefinedFunction);
}

/*-------------------------------------------------------------------------- */
