/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

#include "tcpConn.h"
#include "sockets.h"

#include <fcntl.h>

typedef struct tcpConn
{ nConn nc;
  FILE* remote;
} tcpConn;

/*-------------------------------------------------------------------------- */

nConnStatus 
tcpConnInitialize(nConn* c)
{
  /* don't need to do anything */
  return(nConnStatusOK);
}

/*-------------------------------------------------------------------------- */

nConnStatus 
tcpConnRead(nConn* conn,unsigned char* val)
/* asynchronous read */
{
  tcpConn* c = (tcpConn*)conn;
  long nextChar;
  if (c->nc.lastOpWasRead == false)
   { rewind(c->remote);
     c->nc.lastOpWasRead = true;
   }
  nextChar = getc(c->remote);
  if (nextChar == EOF)
   { if (feof(c->remote))
       return(nConnStatusReadEOF);
     else if (ferror(c->remote))
       return(nConnStatusReadFailed);
     else
       return(nConnStatusReadNoData);
   }
  else
   { if (val != NULL)
       *val = nextChar & 0xFF;
     return(nConnStatusOK);
   }
}

/*-------------------------------------------------------------------------- */

nConnStatus 
tcpConnPeek(nConn* conn,unsigned char* val)
/* asynchronous peak */
{
  tcpConn* c = (tcpConn*)conn;
  long nextChar;
  if (c->nc.lastOpWasRead == false)
   { rewind(c->remote);
     c->nc.lastOpWasRead = true;
   }
  nextChar = getc(c->remote);
  if (nextChar == EOF)
   { if (feof(c->remote))
       return(nConnStatusReadEOF);
     else if (ferror(c->remote))
       return(nConnStatusReadFailed);
     else
       return(nConnStatusReadNoData);
   }
  else
   { if (val != NULL)
      { *val = nextChar & 0xFF;
	if (ungetc(nextChar,c->remote) == EOF)
	  return(nConnStatusReadEOF);
      }
     return(nConnStatusOK);
   }
}

/*-------------------------------------------------------------------------- */

nConnStatus 
tcpConnWrite(nConn* conn,unsigned char val)
{
  long res;
  tcpConn* c = (tcpConn*)conn;

  if (c->nc.lastOpWasRead)
   { rewind(c->remote);
     c->nc.lastOpWasRead = false;
   }
  
#ifdef __alpha /* this is needed for osf1/v2, and maybe for hpux 8.0 as well */
  { char buf[1];
    buf[0] = val;
    res = write(fileno(c->remote),buf,1L);
  }
#else
  res = putc(val,c->remote);
#endif

  if (res == EOF)
    return(nConnStatusWriteFailed);
  else
    return(nConnStatusOK);
}

/*-------------------------------------------------------------------------- */

nConnStatus 
tcpConnFlush(nConn* conn)
{
  tcpConn* c = (tcpConn*)conn;
  c->nc.lastOpWasRead = false;
  fflush(c->remote);
  return(nConnStatusOK);
}

/*-------------------------------------------------------------------------- */

nConnStatus 
tcpConnFree(nConn* conn)
{
  tcpConn* c = (tcpConn*)conn;
  closeConnectionToServer(c->remote);
  safeFree(conn->name)
  safeFree(c);
  return(nConnStatusOK);
}

/*-------------------------------------------------------------------------- */

nConnStatus 
tcpConnGetId(nConn* conn,char** netId,char** name)
/* fill in the netId and name with information about the entity on the
   other side of the connection.  In this case the ip number and host name.
   We malloc the fields.
 */
{
  tcpConn* c = (tcpConn*)conn;
  struct sockaddr_in source;
  long sourcelen;

  if (netId != NULL)
    *netId = NULL;
  if (name != NULL)
    *name = NULL;
     
  sourcelen = sizeof(struct sockaddr_in);

  /* get information on the machine on the other end of the connection
     (our peer).  note - getpeername returns 0 if successful
   */
  if (!getpeername(fileno(c->remote),(struct sockaddr*)&source,&sourcelen))
   { struct hostent* peer = NULL;

     /* we have information, find the peer's hostname */
     peer = gethostbyaddr((char*)&source.sin_addr,4,AF_INET);
     if (peer != NULL && name != NULL)
       *name = safeStrdup(peer->h_name);
     
     /* and internet address */
     if (netId != NULL)
       *netId = safeStrdup(inet_ntoa(source.sin_addr));

     /* if there was no hostname default to the address */
     if (name != NULL && *name == NULL)
       *name = safeStrdup(*netId);

     return(nConnStatusOK);
   }
  else
   { return(nConnStatusNoIDAvailable);
   }
}

/*-------------------------------------------------------------------------- */

nConnStatus
tcpConnSharedOpen(tcpConn* tc,nConn** conn)
{
  nConn* nc;

  nc = (nConn*)tc;

  if (tc->remote != NULL)
   { /* make its reads non blocking */
#ifdef never
/* XXX
     long fd = fileno(tc->remote);
     if (fcntl(fd,F_SETFL,FNDELAY) == -1)
      { /* XXX handle error */
	return(nConnStatusOpenFailed);
      }
     else
#endif
      { *conn = (nConn*)tc;
	return(nConnStatusOK);
      }
   }
  else
   { /* XXX handle error */
     return(nConnStatusOpenFailed);
   }
}

/*-------------------------------------------------------------------------- */

nConnStatus 
tcpConnOpen(nConn** conn,long numArgs,va_list ap)
/* make a tcp connection to the specified server and port.  The connection
   does not block on reads
 */
{
  char* server;
  long port;
  tcpConn* tc;
  nConnStatus ns;

  if (numArgs != 2)
    return(nConnStatusBadArguments);

  server = (char*)va_arg(ap,char*);
  port = (long)va_arg(ap,long);

  tc = (tcpConn*)safeMalloc(sizeof(tcpConn));
  tc->remote = connectToServer(server,port);

  /* If failed, clean up memory. */
  if ((ns = tcpConnSharedOpen(tc,conn)) != nConnStatusOK)
    safeFree(tc);

  return(ns);
}

/*-------------------------------------------------------------------------- */

nConnStatus 
tcpConnWaitForConnection(nConn** conn,long numArgs,va_list ap)
{ 
  FILE* file;
  long socket;
  long port;
  tcpConn* tc;
  nConnStatus ns;

  if (numArgs != 1)
    return(nConnStatusBadArguments);

  port = (long)va_arg(ap,long);

  openServer(port,&socket,16L*1024L);
  acceptClientConnection(socket,&file);
  closeServer(socket);

  tc = (tcpConn*)safeMalloc(sizeof(tcpConn));
  tc->remote = file;

  /* If failed, clean up memory. */
  if ((ns = tcpConnSharedOpen(tc,conn)) != nConnStatusOK)
    safeFree(tc);

  return(ns);
}

/*-------------------------------------------------------------------------- */
