/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

#include "conn1988Cli.h"
#include "wprot.h"
#include "waislog.h"
#include "wmessage.h"

/*---------------------------------------------------------------------------*/
/* INIT */
/*---------------------------------------------------------------------------*/

appConnStatus 
z39501988InitReqWrite(appConn* conn,
		      char*    user,    
		      char*    password,
		      char*    clientInfo,
		      any*     userInfo,
		      list*    diags)
{
  InitAPDU* init = NULL;
  char* start = conn->buf + HEADER_LENGTH;
  char* end;
  long size;
  char* message;

  message = safeMalloc(strlen(user) + strlen(password) + strlen(clientInfo)+1);
  strcpy(message,user);
  strcat(message,password);
  strcat(message,clientInfo);

  /* construct an init, requesting a buffer size = to our internal buffer */
  init = makeInitAPDU(true,false,false,false,false,conn->realBufSz,
		      conn->realBufSz,message,defaultImplementationID(),
		      defaultImplementationName(),
		      defaultImplementationVersion(),NULL,NULL);
  safeFree(message);

  /* write it to the buffer (note don't know negotiated BufSz yet */
  size = conn->realBufSz - HEADER_LENGTH;;
  end = writeInitAPDU(init,start,&size);
  size = end - start;

  /* clean up */
  freeInitAPDU(init);

  /* send it to the remote system
     and create a diagnostic if a network problem was encountered. */
  z39501988WriteMessage(conn->net,conn->buf,size,diags);

  return(appConnStatusOK);
}

/*---------------------------------------------------------------------------*/

appConnStatus 
z39501988InitRespRead(appConn* conn,  
		      any**    userInfo,
		      list*    diags)
{
  InitResponseAPDU* reply = NULL;
  char* start = conn->buf + HEADER_LENGTH;

  /* decode it */
  readInitResponseAPDU(&reply,start);

  if (reply != NULL) /* see what it says */
   { if (reply->Result)
      { /* set the new max buffer size */
	conn->negotiatedBufSz = reply->MaximumRecordSize;
	/* when forwarding, if this size is > serverBufferSize (the size we 
	   negotiated with our client), we may have a problem.  That's why 
	   we try to negotiate a small size with our client up front.
	 */
      }
     else 
       { *diags = collecting(*diags,
			     appConnDiagMake(appConnDiagAccessDenied,""));
       }
     freeWAISInitResponse((WAISInitResponse*)reply->UserInformationField);
   }
  else
    { *diags = collecting(*diags,
			  appConnDiagMake(appConnDiagConnErrRead,
					  "unable to read initialization response"));
    }

  /* clean up */
  freeInitResponseAPDU(reply);
  
  return(appConnStatusOK);
}

/*---------------------------------------------------------------------------*/
/* SEARCH */
/*---------------------------------------------------------------------------*/

boolean
parseQry(appConnQry* qry,char** seedWords,DocObj*** docs)
/* parse the query, the only thing we support is seed words and
   ORs of rel docs.  Returns true if this qry is OK.
 */
{
  if (qry->type == qryLeafFreeText)
   { if (*seedWords == NULL)
      { *seedWords = anyToString(qry->term);
	return(true);
      }
     else
       return(false); /* seed words already set, not allowed */
   }
  else if (qry->type == qryLeafRFRecId)
   {
     appConnQryRF* rf = qry->rf;
     long i = 0L;

     for(i=0; ((*docs) != NULL) && ((*docs)[i] != NULL); i++);
     *docs = (DocObj**)safeRealloc(*docs,(size_t)(sizeof(DocObj*)*(i+2)));

     if (rf->lengthInBytes == MAXLONG)
       (*docs)[i] = makeDocObjUsingWholeDocument(duplicateAny(rf->recId),
						 safeStrdup(rf->varId));
     else
       (*docs)[i] = makeDocObjUsingBytes(duplicateAny(rf->recId),
					 safeStrdup(rf->varId),
					 rf->startByte,rf->lengthInBytes);
     (*docs)[i+1] = NULL;
   }
  else if (qry->type == qrySubTree)
   { if (parseQry(qry->arg1,seedWords,docs) == false)
       return(false);
     if (parseQry(qry->arg2,seedWords,docs) == false)
       return(false);
     return(true);
   }
}

appConnStatus 
z39501988SearchReqWrite(appConn*    conn,  
			long        maxHits,
			long        hitsRequested, 
			long        rankingScheme, 
			char*       dbName, 
			appConnQry* qry,
			any*        userInfo,
			list*       diags)
{
  char* dbNames[2];
  DocObj** docs = NULL;
  char* seedWords = NULL;
  long i = 0;
  char* start = conn->buf + HEADER_LENGTH;
  char* end = NULL;
  long size;
  WAISSearch* waisSearch = NULL;
  SearchAPDU* z3950Search = NULL;
  appConn1988* zConn = (appConn1988*)conn;

  dbNames[0] = dbName;
  dbNames[1] = NULL;

  if (!parseQry(qry,&seedWords,&docs))
    *diags = collecting(*diags,
			appConnDiagMake(appConnDiagUnsupportedSearch,
					"search not supported in Z39.50-V1"));

  waisSearch = makeWAISSearch(seedWords,docs,NULL,DF_INDEPENDENT,NULL,
			      NULL,maxHits);
  z3950Search = makeSearchAPDU(30L,5000L,30L,true,"",dbNames,
			       QT_RelevanceFeedbackQuery,NULL,NULL,
			       waisSearch);
  size = conn->negotiatedBufSz - HEADER_LENGTH;
  end = writeSearchAPDU(z3950Search,start,&size);
  size = end - start;
  freeWAISSearch(waisSearch);
  freeSearchAPDU(z3950Search);

  zConn->doingSearch = true;

  /* send it to the remote system */
  z39501988WriteMessage(conn->net,conn->buf,size,diags);

  return(appConnStatusOK);
}

/*---------------------------------------------------------------------------*/

void
z39501988DiagsToAppConnDiags(WAISSearchResponse* waisResponse,
			     list*               diags)

{ 
  diagnosticRecord* zdiag = NULL;
  long i;

  if (waisResponse->Diagnostics != NULL)

   { for (i = 0, zdiag = (void *)waisResponse->Diagnostics[i]; 
	  zdiag != NULL; 
	  zdiag = (void *)waisResponse->Diagnostics[++i])

      { long code;
	if (strcmp(zdiag->DIAG,D_PresentRequestOutOfRange) == 0)
	 { if (strcmp(zdiag->ADDINFO,"Bad ChunkType in Request") == 0)
	     code = appConnDiagInvalidChunkType;
           else if (strcmp(zdiag->ADDINFO,"Bad DocID in request") == 0)
	     code = appConnDiagInvalidDocId;
	   else if (strcmp(zdiag->ADDINFO,"Missing DocID in request") == 0)
	     code = appConnDiagNoDocIdSpecified;
	   else if (strcmp(zdiag->ADDINFO,"Request out of range") == 0)
	     code = appConnDiagRequestOutOfRange;
	   else if (strcmp(zdiag->ADDINFO,"Database missing from request")
		    == 0)
	     code = appConnDiagNoDatabaseSpecified;
	   else if (strcmp(zdiag->ADDINFO,
			   "File not present in specified database") == 0)
	     code = appConnDiagDocumentNotFound;
	   else code = appConnDiagServerError;
	 }
	else code = appConnDiagRequestOutOfRange;
	*diags = collecting(*diags,appConnDiagMake(code,zdiag->ADDINFO));
      }
   }
}

/*---------------------------------------------------------------------------*/

appConnStatus 
z39501988SearchRespRead(appConn* conn,
			long*    totalHits, 
			list*    qryTermsUsed,
			list*    hits,
			any**    userInfo,
			list*    diags)
/* note that 1988 has no idea of the difference between elements and 
   variants.  We'll fudge this by making each type a separate element
   with one variant
 */
{
  SearchResponseAPDU* z3950Response = NULL;
  WAISSearchResponse* waisResponse = NULL;
  char* start = conn->buf + HEADER_LENGTH;
  char* s;

  /* decode it */
  readSearchResponseAPDU(&z3950Response,start);

  if (z3950Response != NULL)
    waisResponse = 
      (WAISSearchResponse*)z3950Response->DatabaseDiagnosticRecords;
  if (z3950Response == NULL || waisResponse == NULL)
   { *diags = collecting(*diags,appConnDiagMake(appConnDiagBadPacket,""));
   }
  else /* turn it into something we can give back to our caller */
   { long i;
     for (i = 0; waisResponse != NULL && waisResponse->DocHeaders != NULL; i++)
      { WAISDocumentHeader* doc = waisResponse->DocHeaders[i];
	long j;
	appConnHeadline* hl = NULL;
	list elements = NULL;
	date* theDate;

	if (doc == NULL)
	  break;
	
	theDate = YYYYMMDDToDate(doc->Date);

	for (j = 0; doc->Types != NULL; j++)
	 { char* type = doc->Types[j];
	   appConnVariant* v;
	   appConnElement* e;
	   list vl = NULL;

	   if (type == NULL)
	     break;

	   v = appConnMakeVariant(type,type,NULL,NULL,
				  doc->DocumentLength,0,appConnUnknown);
	   vl = collecting(vl,v);
	   e = appConnMakeElement(type,NULL,vl,NULL);
	   elements = collecting(elements,e);
	 }

	hl = appConnMakeHeadline(doc->Headline,NULL,theDate,i,doc->Score,
				 doc->DocumentID,NULL,elements);
	safeFree(theDate);

	*hits = collecting(*hits,hl);
      }

     z39501988DiagsToAppConnDiags(waisResponse,diags);
   }

  /* Parse the seed words into a list. */
  if (((s = waisResponse->SeedWordsUsed) != NULL) &&
      (qryTermsUsed != NULL))
   {
     while (true)
      { 
	long n = strcspn(s," ");
	char* term = safeStrndup(s,n+1);

	*qryTermsUsed = collecting(*qryTermsUsed,term);

	/* check for the end of the string */
	if (n == strlen(s)) break;

	/* go to the next term and skip over the space */
	s += n+1;
      }
   }

  /* in 1988 we always get back all the hits */
  *totalHits = length(*hits);

  freeWAISSearchResponse(waisResponse);
  freeSearchResponseAPDU(z3950Response);

  return(appConnStatusOK);
}

/*---------------------------------------------------------------------------*/

appConnStatus 
z39501988NextHeadlinesReqWrite(appConn* conn,
			       long     startingFromHit, 
			       long     hitsRequested,
			       any*     userInfo,
			       list*    diags)
{
  /* this is a nop in the 1988 protocol, since there are no presents */
  return(appConnStatusUndefinedFunction);
}

/*---------------------------------------------------------------------------*/

appConnStatus 
z39501988NextHeadlinesRespRead(appConn* conn,
			       list*    hits,
			       any**    userInfo,
			       list*    diags)
{
  *hits = NULL;
  *userInfo = NULL;
  *diags = collecting(*diags,
		      appConnDiagMake(appConnDiagNoMoreHeadlines,
				      "in Z3950-V1"));
  return(appConnStatusUndefinedFunction);  
}

/*---------------------------------------------------------------------------*/
/* RETRIEVAL */
/*---------------------------------------------------------------------------*/

appConnStatus 
z39501988RetrieveReqWrite(appConn* conn,
			  long     retrievalType,
			  long     resSetPos, 
			  any*     recId, 
			  char*    docId,
			  char*    elemId, 
			  char*    varId, 
			  char*    dbName,
			  long     startByte,
			  long     lengthInBytes,
			  any*     userInfo,
			  list*    diags)
{
  SearchAPDU* z3950Search = NULL;
  char* dbNames[2];
  char* elementNames[3]; 
  DocObj* docs[2];
  any* query = NULL;
  long i = 0;
  char* start = conn->buf + HEADER_LENGTH;
  char* end = NULL;
  long size;
  appConn1988* zConn = (appConn1988*)conn;

  if (retrievalType != appConnRetrieveByRecId)
    return(appConnStatusBadArguments);

  /* format the request in z3950 */
  dbNames[0] = dbName;
  dbNames[1] = NULL;
  elementNames[0] = " ";
  elementNames[1] = ES_DocumentText;
  elementNames[2] = NULL;

  if (lengthInBytes == MAXLONG)
   { docs[0] = makeDocObjUsingWholeDocument(duplicateAny(recId),
					    safeStrdup(varId));
   }
  else
   { docs[0] = makeDocObjUsingBytes(duplicateAny(recId),safeStrdup(varId),
				    startByte,startByte + lengthInBytes);
   }
  docs[1] = NULL;

  query = makeWAISTextQuery(docs);   
  freeDocObj(docs[0]);

  z3950Search = makeSearchAPDU(30L,5000L,30L,true,"",
			       dbNames,QT_TextRetrievalQuery,
			       elementNames, NULL,query);
  size = conn->negotiatedBufSz - HEADER_LENGTH;
  end = writeSearchAPDU(z3950Search,start,&size);
  size = end - start;
  freeAny(query);
  freeSearchAPDU(z3950Search);

  zConn->doingSearch = false;

  /* send it to the remote system */
  z39501988WriteMessage(conn->net,conn->buf,size,diags);

  return(appConnStatusOK);
}

/*---------------------------------------------------------------------------*/

appConnStatus 
z39501988RetrieveRespRead(appConn* conn,
			  char**   data,
			  long*    dataLength,
			  any**    userInfo,
			  list*    diags)
{
  SearchResponseAPDU* z3950Response = NULL;
  WAISSearchResponse* waisResponse = NULL;
  char* start = conn->buf + HEADER_LENGTH;

  /* decode it */
  readSearchResponseAPDU(&z3950Response,start);

  if (z3950Response)
    waisResponse = 
      (WAISSearchResponse*)z3950Response->DatabaseDiagnosticRecords;
  if (z3950Response == NULL || waisResponse == NULL)
    *diags = collecting(*diags,appConnDiagMake(appConnDiagBadPacket,""));
  else /* turn it into something we can give back to our caller */
   { if (waisResponse->Text != NULL)
      { *dataLength = waisResponse->Text[0]->DocumentText->size;
	*data = waisResponse->Text[0]->DocumentText->bytes;
	/*	if (strcmp(type,"WSRC") == 0)
		modifyWSRCdata(data,dataLength,serverHost,serverPort);
		XXX - required for forwarder
		*/
	waisResponse->Text[0]->DocumentText->bytes = NULL;
      }

     		    
       z39501988DiagsToAppConnDiags(waisResponse,diags);
       freeWAISSearchResponse(waisResponse);
       freeSearchResponseAPDU(z3950Response);
     }

  return(appConnStatusOK);  
}

/*---------------------------------------------------------------------------*/
