/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

#include "conn1988Mgr.h"
#include "waislog.h"
#include "wmessage.h"

/*---------------------------------------------------------------------------*/
/* Utilities */
/*-------------------------------------------------------------------------- */

nConnStatus
z39501988WriteMessage(nConn* net,
		      char* buf,
		      long size,
		      list* diags)
/* write the message.  This function also writes the wais packet header 
   which is essencially our transport layer
 */
{
  long i;
  WAISMessage header;		
  char lengthArray[11];
  nConnStatus res;

  writeWAISPacketHeader(buf,size,Z3950,"forwarder",
			NO_COMPRESSION,NO_ENCODING,HEADER_VERSION);

  /* send it to the remote server, and wait for a response */
  for (i = 0; i < size + HEADER_LENGTH; i++)
   { res = nConnWrite(net,buf[i]);
     if (res != nConnStatusOK)
       return(res);
   }

  /* waislog(WLOG_LOW,WLOG_INFO,"response packet size = %d",
     size+HEADER_LENGTH); */

  res = nConnFlush(net);		

  if (res != nConnStatusOK)
    *diags
      = collecting(*diags,
		   appConnDiagMake(appConnDiagConnErrWrite,
				   "error writing %ld bytes to connection",
				   size));
  return(res);
}

/*---------------------------------------------------------------------------*/

appConnStatus 
z39501988Open(appConn** conn,char* connType,nConn* net,
	      char* buffer,long bufSz,long timeOutSecs,
	      long numArgs,...)
/* nothing special to do here */
{
  appConn* c;

  *conn = c = (appConn*)safeMalloc(sizeof(appConn1988));

  c->name = safeStrdup(connType);
  c->net = net;
  c->buf = buffer;
  c->realBufSz = bufSz;
  c->negotiatedBufSz = bufSz;
  c->timeOutSecs = timeOutSecs;
  
  return(appConnStatusOK);
}

/*---------------------------------------------------------------------------*/

appConnStatus z39501988Free(appConn* conn)
{
  appConn1988* zConn = (appConn1988*)conn;

  safeFree(zConn);
  return(appConnStatusOK);
}

/*---------------------------------------------------------------------------*/
/* PDU dispatching */
/*---------------------------------------------------------------------------*/

appConnStatus 
z39501988PduInfo (appConn* conn,
		  long*    pduType,
		  long*    packetLength,
		  list*    diags)
/* for 1988, there is a 25 byte header which encodes the total packet
   length.  We figure that out and read the rest of the bytes, then
   parse the first few to determine the apdu.
 */
{
  appConn1988* zConn = (appConn1988*)conn;
  time_t startTime = time(NULL);
  char* start = conn->buf + HEADER_LENGTH;
  pdu_type pdu;
  
  if (packetLength != NULL)
    *packetLength = appConnUnknownPacketLength; /* the default */ 

  if (zConn->readSoFar == zConn->totalLength && 
      zConn->transportDecodingDone == true)
   { /* we finished reading the previous batch, and are being asked to 
	read a new batch, so we need to initialize things
      */
     zConn->totalLength = 0;
     zConn->readSoFar = 0;
     zConn->transportDecodingDone = false;
   }

  /* read the header if necessary */
  for (zConn->readSoFar; zConn->readSoFar < HEADER_LENGTH;)
   { unsigned char nextChar;
     nConnStatus err;
     err = nConnRead(conn->net,&nextChar);
     if (err == nConnStatusOK)
      { conn->buf[zConn->readSoFar] = nextChar;
	zConn->readSoFar++;
      }
     else if (err == nConnStatusReadFailed)
      { *diags = collecting(*diags,appConnDiagMake(appConnDiagConnErrRead,""));

	return(appConnStatusOK);
      }
     else if (err == nConnStatusReadEOF)
      { *diags = collecting(*diags,appConnDiagMake(appConnDiagConnClosed,""));
	return(appConnStatusOK);
      }
     else if (err == nConnStatusReadNoData)
       ; /* no data yet, try again latter */
     if (timeOut(startTime,conn->timeOutSecs))
       return(appConnStatusNotFinished);       
   }

  /* parse the header if we have not done so */
  if (zConn->readSoFar == HEADER_LENGTH)
   { WAISMessage header;		
     char lengthArray[11];
     readWAISPacketHeader(conn->buf,&header);
     if (header.hdr_vers > HEADER_VERSION)
      { *diags = collecting(*diags,
			    appConnDiagMake(appConnDiagServerError,
					    "incompatable header version %c expected %c",
					    header.hdr_vers,HEADER_VERSION));
	return(appConnStatusOK);
      }
     strncpy(lengthArray,header.msg_len,10);
     lengthArray[10] = '\0';
     zConn->totalLength = HEADER_LENGTH + atol(lengthArray);
     zConn->encoding = header.encoding;
   }

  /* by the time we get here, we know how long the packet is */
  if (packetLength != NULL)
    *packetLength = zConn->totalLength;

  /* read the rest of the message if necessary */
  for (zConn->readSoFar; zConn->readSoFar < zConn->totalLength;)
   { unsigned char nextChar;
     nConnStatus err;
     err = nConnRead(conn->net,&nextChar);
     if (err == nConnStatusOK)
      { conn->buf[zConn->readSoFar] = nextChar;
	zConn->readSoFar++;
      }
     else if (err == nConnStatusReadFailed || err == nConnStatusReadEOF)
      { *diags = collecting(*diags,appConnDiagMake(appConnDiagConnErrRead,""));
	return(appConnStatusOK);
      }
     else if (err == nConnStatusReadNoData)
       ; /* no data yet, try again latter */
     if (timeOut(startTime,conn->timeOutSecs))
       return(appConnStatusNotFinished);       
   }

  /* undo the transport encoding if necessary */
  if (zConn->transportDecodingDone == false)
   { /* decode it */
     transportDecode(zConn->encoding,start,
		     &zConn->totalLength - HEADER_LENGTH);
     zConn->transportDecodingDone = true;

     /* finally find out what kind of data we've got */
     pdu = peekPDUType(start);
     switch (pdu)
      { case initAPDU:
	  *pduType = appConnInitReq;
	  break;
	case initResponseAPDU:
	  *pduType = appConnInitResp;
	  break;
	case searchAPDU:
	 { /* in 1988, there is no present, so everything is done
	      with a search.  That means we have to actually decode part
	      of the message here to figure out if a search or present
	      was being done.
	    */
	   if (readSearchAPDU(&(zConn->searchPDU),
			      conn->buf+HEADER_LENGTH) == NULL)
	    { *diags = collecting(*diags,
				  appConnDiagMake(appConnDiagBadPacket,""));
	      return(appConnStatusOK);
	    }
	   else 
	    { if (strcmp(zConn->searchPDU->QueryType,
			 QT_RelevanceFeedbackQuery) == 0) 
	       { *pduType = appConnSearchReq;
		 return(appConnStatusOK);
	       }
              else if (strcmp(zConn->searchPDU->QueryType,
			      QT_TextRetrievalQuery) == 0) 
	       { *pduType = appConnRetrieveReq;
		 return(appConnStatusOK);
	       }
	      else
	       { *pduType = appConnUnknown;
		 return(appConnStatusOK);
	       }
	    }

	   break;
	 }
	case searchResponseAPDU:
	 { /* in 1988, there is no present, so everything is done
	      with a search.  That means we have to remember what we were
	      doing so that we can know what to expect
	    */
	   if (zConn->doingSearch)
	     *pduType = appConnSearchResp;
	   else
	     *pduType = appConnRetrieveResp;
	   break;
	 }
	case presentAPDU:
	  /* we never get here, since there is no present in 1988 */
	  break;
	case presentResponseAPDU:
	  /* we never get here, since there is no present in 1988 */
	  break;
	default:
	  *pduType = appConnUnknown;
	  break;
      }
     return(appConnStatusOK);
   }

  return(appConnStatusOK);
}

/*---------------------------------------------------------------------------*/
