#include <ctype.h>
#include "docid-io.h"
#include "irfileio.h"

/*---------------------------------------------------------------------------*/

long 
WriteDocID(DocID* doc,FILE* file)
{
  WriteNewline(file);
  WriteStartOfStruct("doc-id",file);
  if (doc->originalServer != NULL)
   { WriteNewline(file);
     WriteSymbol(":original-server",file); 
     WriteAny(doc->originalServer,file);
   }
  if (doc->originalDatabase != NULL)
   { WriteNewline(file);
     WriteSymbol(":original-database",file); 
     WriteAny(doc->originalDatabase,file);
   }
  if (doc->originalLocalID != NULL)
   { WriteNewline(file);
     WriteSymbol(":original-local-id",file); 
     WriteAny(doc->originalLocalID,file);
   }
  if (doc->distributorServer != NULL)
   { WriteNewline(file);
     WriteSymbol(":distributor-server",file); 
     WriteAny(doc->distributorServer,file);
   }
  if (doc->distributorDatabase != NULL)
   { WriteNewline(file);
     WriteSymbol(":distributor-database",file); 
     WriteAny(doc->distributorDatabase,file);
   }
  if (doc->distributorLocalID != NULL)
   { WriteNewline(file);
     WriteSymbol(":distributor-local-id",file); 
     WriteAny(doc->distributorLocalID,file);
   }
  WriteNewline(file);
  WriteSymbol(":copyright-disposition",file);
  WriteLong(doc->copyrightDisposition,file);
  return(WriteEndOfStruct(file));
}

/*---------------------------------------------------------------------------*/

long 
ReadDocID(DocID* doc,FILE* file)
{
  char temp_string[MAX_SYMBOL_SIZE];

  if (CheckStartOfStruct("doc-id",file) == false)
    return(false);

  doc->originalServer = NULL;
  doc->originalDatabase = NULL;
  doc->originalLocalID = NULL;
  doc->distributorServer = NULL;
  doc->distributorDatabase = NULL;
  doc->distributorLocalID = NULL;
  doc->copyrightDisposition = COPY_WITHOUT_RESTRICTION;
	
  while(true)
   { long result = ReadSymbol(temp_string,file,(long)MAX_SYMBOL_SIZE);
		
     if (result == false)
       return(false);
     if (result == END_OF_STRUCT_OR_LIST)
       break;
     if (strcmp(temp_string,":original-server") == 0)
      { any* an_any = (any*)safeMalloc(sizeof(any));
	ReadAny(an_any,file);
	doc->originalServer = an_any;
      }
     else if (strcmp(temp_string,":original-database") == 0)
      { any* an_any = (any*)safeMalloc(sizeof(any));
	ReadAny(an_any,file);
	doc->originalDatabase = an_any;
      }
     else if (strcmp(temp_string,":original-local-id") == 0)
      { any* an_any = (any*)safeMalloc(sizeof(any));
	ReadAny(an_any,file);
	doc->originalLocalID = an_any;
      }
     else if (strcmp(temp_string,":distributor-server") == 0)
      { any* an_any = (any*)safeMalloc(sizeof(any));
	ReadAny(an_any,file);
	doc->distributorServer = an_any;
      }
     else if (strcmp(temp_string,":distributor-database") == 0)
     { any* an_any = (any*)safeMalloc(sizeof(any));
       ReadAny(an_any,file);
       doc->distributorDatabase = an_any;
     }
    else if (strcmp(temp_string,":distributor-local-id") == 0)
     { any* an_any = (any*)safeMalloc(sizeof(any));
       ReadAny(an_any,file);
       doc->distributorLocalID = an_any;
     }
    else if (strcmp(temp_string,":copyright-disposition") == 0)
      ReadLong(file,&doc->copyrightDisposition);
    else
      SkipObject(file);
   }
  return(true);
}

/*---------------------------------------------------------------------------*/

#define DT_OriginalServer 	((data_tag)1)
#define DT_OriginalDatabase 	((data_tag)2)
#define DT_OriginalLocalID	((data_tag)3)
#define DT_DistributorServer	((data_tag)4)
#define DT_DistributorDatabase	((data_tag)5)
#define DT_DistributorLocalID	((data_tag)6)
#define DT_CopyrightDispostion	((data_tag)7)

DocID*
docIDFromAny(any* rawDocID)
{
  DocID* docID;
  char* buf;
  
  TESTDOC(rawDocID);

  buf = rawDocID->bytes;
  docID = makeDocID();

  while ((buf - rawDocID->bytes) < rawDocID->size)
   { data_tag tag = peekTag(buf);
     switch (tag)
      { case DT_OriginalServer:
	  buf = readAny(&(docID->originalServer),buf);
	  break;
	case DT_OriginalDatabase:
	  buf = readAny(&(docID->originalDatabase),buf);
	  break;
	case DT_OriginalLocalID:
	  buf = readAny(&(docID->originalLocalID),buf);
	  break;
	case DT_DistributorServer:
	  buf = readAny(&(docID->distributorServer),buf);
	  break;
	case DT_DistributorDatabase:
	  buf = readAny(&(docID->distributorDatabase),buf);
	  break;
	case DT_DistributorLocalID:
	  buf = readAny(&(docID->distributorLocalID),buf);
	  break;
	case DT_CopyrightDispostion:
	  buf = readNum(&(docID->copyrightDisposition),buf);
	  break;
	default:
	  freeDocID(docID);
	  return(NULL);
	};
   }
  
  return(docID);
}

/*---------------------------------------------------------------------------*/

any*
anyFromDocID(DocID* docID)
{
  any* rawDocID = NULL;
  char* buf = NULL;
  char* data = NULL;
  long size,bytesLeft;
  
  if (docID == NULL)
    return(NULL);
    
  size = writtenAnySize(DT_OriginalServer,docID->originalServer);
  size += writtenAnySize(DT_OriginalDatabase,docID->originalDatabase);
  size += writtenAnySize(DT_OriginalLocalID,docID->originalLocalID);
  size += writtenAnySize(DT_DistributorServer,docID->distributorServer);
  size += writtenAnySize(DT_DistributorDatabase,docID->distributorDatabase);
  size += writtenAnySize(DT_DistributorLocalID,docID->distributorLocalID);
  size += writtenNumSize(DT_CopyrightDispostion,docID->copyrightDisposition);
  
  data = safeMalloc((size_t)(sizeof(char) * size));
  
  buf = data;
  bytesLeft = size;
  
  buf = writeAny(docID->originalServer,DT_OriginalServer,buf,&bytesLeft);
  buf = writeAny(docID->originalDatabase,DT_OriginalDatabase,buf,&bytesLeft);
  buf = writeAny(docID->originalLocalID,DT_OriginalLocalID,buf,&bytesLeft);
  buf = writeAny(docID->distributorServer,DT_DistributorServer,buf,
                 &bytesLeft);
  buf = writeAny(docID->distributorDatabase,DT_DistributorDatabase,
                 buf,&bytesLeft);
  buf = writeAny(docID->distributorLocalID,DT_DistributorLocalID,buf,
                 &bytesLeft);
  buf = writeNum(docID->copyrightDisposition,DT_CopyrightDispostion,
                 buf,&bytesLeft);
  
  rawDocID = makeAny(size,data);
  
  return(rawDocID);
}

/*---------------------------------------------------------------------------*/
