#include <ctype.h>
#include "docid.h"
#include "cutil.h"
#include "cdialect.h"

/*---------------------------------------------------------------------------*/

any* 
GetServer(DocID* doc)
{
  TESTDOC(doc);
  if (doc->distributorServer != NULL)
    return(doc->distributorServer);
  else 
    return(doc->originalServer);
}

/*---------------------------------------------------------------------------*/

any* 
GetDatabase(DocID* doc)
{
  TESTDOC(doc);
  if (doc->distributorDatabase != NULL)
    return(doc->distributorDatabase);
  else 
    return(doc->originalDatabase);
}

/*---------------------------------------------------------------------------*/

any* 
GetLocalID(DocID* doc)
{
  TESTDOC(doc);
  if (doc->distributorLocalID != NULL)
    return(doc->distributorLocalID);
  else 
    return(doc->originalLocalID);
}

/*---------------------------------------------------------------------------*/

long 
GetCopyrightDisposition(DocID* doc)
{
  if (doc)
    return(doc->copyrightDisposition);
  else 
    return(0);
}

/*---------------------------------------------------------------------------*/

static boolean 
safeCmp(any* a1,any* a2)
{
		
  if (a1 == NULL && a2 == NULL)
    return(true);
  else if (a1 == NULL || a2 == NULL)
    return(false);
  else if (a1->size != a2->size)
    return(false);
  else if (memcmp(a1->bytes,a2->bytes,(size_t)a1->size) == 0)
    return(true);
  else
    return(false);
}

/*---------------------------------------------------------------------------*/

boolean 
cmpDocIDs(DocID* d1,DocID* d2)
{
  if (safeCmp(d1->originalServer,d2->originalServer) &&
      safeCmp(d1->originalDatabase,d2->originalDatabase) &&
      safeCmp(d1->originalLocalID,d2->originalLocalID) &&
      safeCmp(d1->distributorServer,d2->distributorServer) &&
      safeCmp(d1->distributorDatabase,d2->distributorDatabase) &&
      safeCmp(d1->distributorLocalID,d2->distributorLocalID) &&
      d1->copyrightDisposition == d2->copyrightDisposition)
    return(true);
  else
    return(false); 
}
	
/*---------------------------------------------------------------------------*/

DocID* 
makeDocID(void)
{
  return((DocID*)safeMalloc((size_t)sizeof(DocID)));
}

/*---------------------------------------------------------------------------*/

DocID*
copyDocID(DocID* doc)
{
  DocID* result = NULL;
  if(doc != NULL && (result = makeDocID()) != NULL) 
   { result->originalServer = duplicateAny(doc->originalServer);
     result->originalDatabase = duplicateAny(doc->originalDatabase);
     result->originalLocalID = duplicateAny(doc->originalLocalID);
     result->distributorServer = duplicateAny(doc->distributorServer);
     result->distributorDatabase = duplicateAny(doc->distributorDatabase);
     result->distributorLocalID = duplicateAny(doc->distributorLocalID);
   }
  return(result);
}

/*---------------------------------------------------------------------------*/

void 
freeDocID(DocID* doc)
{
  if (doc) 
   { freeAny(doc->originalServer);
     freeAny(doc->originalDatabase);
     freeAny(doc->originalLocalID);
     freeAny(doc->distributorServer);
     freeAny(doc->distributorDatabase);
     freeAny(doc->distributorLocalID);
     safeFree(doc);
   }
}

/*---------------------------------------------------------------------------*/

