#include "transprt.h"
#include "wmessage.h" 

/*---------------------------------------------------------------------*/

static char alphaH[] = 
  "0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnop";

/*---------------------------------------------------------------------*/

static void hexCodeH _AP((char* data,long* len));

static void
hexCodeH(char* data,long* len)
{
  long i, j;
  
  switch (*len % 3)
    {
    case 2:
      i = *len; j = (*len+1)/3 * 4 - 1; *len = j + 1;
      data[j]   = alphaH[64];
      data[j-1] = alphaH[data[i-1] & 0x0f];
      data[j-2] = alphaH[(data[i-2] & 0xc0) >> 2 | (data[i-1] & 0xf0) >> 4];
      data[j-3] = alphaH[data[i-2] & 0x3F];
      break;
    case 1:
      i = *len + 1; j = (*len+2)/3 * 4 - 1; *len = j + 1;
      data[j]   = alphaH[64];
      data[j-1] = alphaH[64];
      data[j-2] = alphaH[(data[i-2] & 0xc0) >> 2];
      data[j-3] = alphaH[data[i-2] & 0x3F];
      break;
    default:
      i = *len + 2; j = (*len)/3 * 4 + 3; *len = j - 3;
      break;
    }
  
  for (i -= 3, j -= 4; i > 0; i -= 3, j -= 4)
    { 
      data[j]   = alphaH[data[i] & 0x3F];
      data[j-1] = alphaH[(data[i] & 0xc0) >> 2 | (data[i-1] & 0x0f)];
      data[j-2] = alphaH[(data[i-2] & 0xc0) >> 2 | (data[i-1] & 0xf0) >> 4];
      data[j-3] = alphaH[data[i-2] & 0x3F];
    }
}

/*---------------------------------------------------------------------*/

static void
hexDecodeH(char* data,long* len)
{
  unsigned char b0, b1, b2, b3;
  long i, j;
  
  for (i = 3, j = 4; j < *len; i += 3, j += 4)
    {
      b0 = (data[j-4] & 0x07F) - alphaH[0];
      b1 = (data[j-3] & 0x07F) - alphaH[0];
      b2 = (data[j-2] & 0x07F) - alphaH[0];
      b3 = (data[j-1] & 0x07F) - alphaH[0];
      data[i-3] = b0 | (b1 & 0x30) << 2;
      data[i-2] = (b1 & 0x0f) << 4 | (b2 & 0x0f);
      data[i-1] = b3 | (b2 & 0x30) << 2;
    }

  *len = *len/4 * 3 - 2;
  b0 = (data[j-4] & 0x07F) - alphaH[0];
  b1 = (data[j-3] & 0x07F) - alphaH[0];
  b2 = (data[j-2] & 0x07F) - alphaH[0];
  b3 = (data[j-1] & 0x07F) - alphaH[0];
  data[i-3] = b0 | (b1 & 0x30) << 2;

  if (b2 != 64)    
    {
      data[i-2] = (b1 & 0x0f) << 4 | (b2 & 0x0f);
      (*len)++;
    }

  if (b3 != 64)
    {
      data[i-1] = b3 | (b2 & 0x30) << 2;
      (*len)++;
    }
}

/*---------------------------------------------------------------------*/
 
void
transportCodeH(char* data,long* len)
{ 
  hexCodeH(data,len);
  data[*len] = '\r';
  data[*len + 1] = '\0';
  *len += 1;
}

/*---------------------------------------------------------------------*/

void
transportDecodeH(char* data,long* len)
{ 
  if (*len == 0)
    return;
    
  *len -= 1;
  hexDecodeH(data,len);
}

/*---------------------------------------------------------------------
 Below are the set of transport coding/decoding functions for clients
 dialing through IBM to DowQuest. Unlike the above one that
 uses 65 consecutive ASCII character for the coding/decoding table,
 table entries after charcater 'Z' are shift 6 characters.
 *--------------------------------------------------------------------*/ 

#define LAST_NON_SHIFT_CHAR (90L)
#define SHIFT_OFFSET        (6L)

static char alphaI[] = 
  "0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuv";


static void
hexCodeI(char* data,long* len)
{
  long i, j;
  
  switch (*len % 3)
    {
    case 2:
      i = *len; j = (*len+1)/3 * 4 - 1; *len = j + 1;
      data[j]   = alphaI[64];
      data[j-1] = alphaI[data[i-1] & 0x0f];
      data[j-2] = alphaI[(data[i-2] & 0xc0) >> 2 | (data[i-1] & 0xf0) >> 4];
      data[j-3] = alphaI[data[i-2] & 0x3F];
      break;
    case 1:
      i = *len + 1; j = (*len+2)/3 * 4 - 1; *len = j + 1;
      data[j]   = alphaI[64];
      data[j-1] = alphaI[64];
      data[j-2] = alphaI[(data[i-2] & 0xc0) >> 2];
      data[j-3] = alphaI[data[i-2] & 0x3F];
      break;
    default:
      i = *len + 2; j = (*len)/3 * 4 + 3; *len = j - 3;
      break;
    }
  
  for (i -= 3, j -= 4; i > 0; i -= 3, j -= 4)
    { 
      data[j]   = alphaI[data[i] & 0x3F];
      data[j-1] = alphaI[(data[i] & 0xc0) >> 2 | (data[i-1] & 0x0f)];
      data[j-2] = alphaI[(data[i-2] & 0xc0) >> 2 | (data[i-1] & 0xf0) >> 4];
      data[j-3] = alphaI[data[i-2] & 0x3F];
    }
}

/*---------------------------------------------------------------------*/

static void
hexDecodeI(char* data,long* len)
{
  unsigned char b0, b1, b2, b3;
  long i, j;

  for ( i=0; i < *len ; i++ )
    if ( data[i] > LAST_NON_SHIFT_CHAR ) data[i] -= SHIFT_OFFSET;
  
  for (i = 3, j = 4; j < *len; i += 3, j += 4)
    {
      b0 = (data[j-4] & 0x07F) - alphaI[0];
      b1 = (data[j-3] & 0x07F) - alphaI[0];
      b2 = (data[j-2] & 0x07F) - alphaI[0];
      b3 = (data[j-1] & 0x07F) - alphaI[0];
      data[i-3] = b0 | (b1 & 0x30) << 2;
      data[i-2] = (b1 & 0x0f) << 4 | (b2 & 0x0f);
      data[i-1] = b3 | (b2 & 0x30) << 2;
    }

  *len = *len/4 * 3 - 2;
  b0 = (data[j-4] & 0x07F) - alphaI[0];
  b1 = (data[j-3] & 0x07F) - alphaI[0];
  b2 = (data[j-2] & 0x07F) - alphaI[0];
  b3 = (data[j-1] & 0x07F) - alphaI[0];
  data[i-3] = b0 | (b1 & 0x30) << 2;

  if (b2 != 64)    
    {
      data[i-2] = (b1 & 0x0f) << 4 | (b2 & 0x0f);
      (*len)++;
    }

  if (b3 != 64)
    {
      data[i-1] = b3 | (b2 & 0x30) << 2;
      (*len)++;
    }
}

/*---------------------------------------------------------------------*/
 
void
transportCodeI(char* data,long* len)
{ 
  hexCodeI(data,len);
  data[*len] = '\r';
  data[*len + 1] = '\0';
  *len += 1;
}

/*---------------------------------------------------------------------*/

void
transportDecodeI(char* data,long* len)
{ 
  if (*len == 0)
    return;
    
  *len -= 1;
  hexDecodeI(data,len);
}

/*---------------------------------------------------------------------*/

boolean
transportCode(long encoding,char* data,long* len)
{
  switch (encoding)
   { case NO_ENCODING:
       return(true);
     case HEX_ENCODING:
       transportCodeH(data,len);
       return(true);
     case IBM_HEXCODING:
       transportCodeI(data,len);
       return(true);
     default:
       return(false);
   }
}

/*---------------------------------------------------------------------------*/

boolean
transportDecode(long encoding,char* data,long* len)
{
  switch (encoding)
   { case NO_ENCODING:
       return(true);
     case HEX_ENCODING:
       transportDecodeH(data,len);
       return(true);
     case IBM_HEXCODING:
       transportDecodeI(data,len);
       return(true);
     default:
       return(false);
  }
}

/*---------------------------------------------------------------------------*/




