/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

#include <appConnMgr.h>

#include "diag.h"

#define DIAG_SET_BIB1_000_START 0L
#define DIAG_SET_BIB1_000_LEN   34L
#define DIAG_SET_BIB1_100_START 100L
#define DIAG_SET_BIB1_100_LEN   33L
#define DIAG_SET_BIB1_200_START 200L
#define DIAG_SET_BIB1_200_LEN   32L

#define MAX_DIAG_SET_BIB1_COND (DIAG_SET_BIB1_000_LEN + \
			DIAG_SET_BIB1_100_LEN + \
			DIAG_SET_BIB1_200_LEN)

typedef struct {
  char* sZ3950V2Info;
  long  sAppConnDiagCond;
} tDiag;

tDiag gBib1Diag[MAX_DIAG_SET_BIB1_COND] =
{
/* Diagnostic 0-100 */
 {"unknown condition",appConnDiagProtocolError},
 {"permanent system error",appConnDiagProtocolError},
 {"temporary system error",appConnDiagProtocolError},
 {"unsupported search",appConnDiagUnsupportedSearch},
 {"terms only exclusion (stop) words",appConnDiagProtocolError},
 {"too many argument words",appConnDiagProtocolError},
 {"too many boolean operators",appConnDiagProtocolError},
 {"too many truncated words",appConnDiagProtocolError},
 {"too many incomplete subfields",appConnDiagProtocolError},
 {"truncated words too short",appConnDiagProtocolError},
 {"invalid format for record number (search term)",appConnDiagProtocolError},
 {"too many characters in search statement",appConnDiagProtocolError},
 {"too many records retrieved",appConnDiagProtocolError},
 {"present request out of range",appConnDiagRequestOutOfRange},
 {"system error in presenting records",appConnDiagProtocolError},
 {"record non authorized to be sent intersystem",appConnDiagAccessDenied},
 {"record exceeds preferred message size",appConnDiagProtocolError},
 {"record exceeds maximum record size",appConnDiagProtocolError},
 {"result set not supported as a search term",appConnDiagProtocolError},
 {"only single result set as search term supported",appConnDiagProtocolError},
 {"only ANDing of a single result set as search term supported",appConnDiagProtocolError},
 {"result set exists and replace indicator off",appConnDiagProtocolError},
 {"result set naming not supported",appConnDiagProtocolError},
 {"combination of specified databases not supported",appConnDiagProtocolError},
 {"element set names not supported",appConnDiagProtocolError},
 {"specified element set name not valid for specified database",appConnDiagProtocolError},
 {"only single element set name supported",appConnDiagProtocolError},
 {"result set no longer exists - unilaterally deleted by target",appConnDiagProtocolError},
 {"result set is in use",appConnDiagProtocolError},
 {"one of the specified databases is locked",appConnDiagProtocolError},
 {"specified result set does not exist",appConnDiagProtocolError},
 {"resources exhausted, no results available",appConnDiagProtocolError},
 {"resources exhausted, unpredictable partial results available",appConnDiagProtocolError},
 {"resources exhausted, valid subset of results avaliable",appConnDiagProtocolError},

/* Diagnostic 100-132 */
 {"(unspecified) error",appConnDiagProtocolError},
 {"access control failure",appConnDiagAccessDenied},
 {"challenge required, could not be issued - operation terminated",appConnDiagAccessDenied},
 {"challenge required, could not be issued - record not included",appConnDiagAccessDenied},
 {"challenge failed - record not included",appConnDiagAccessDenied},
 {"terminated at origin request",appConnDiagProtocolError},
 {"no abstract syntaxes agreed to for this record",appConnDiagProtocolError},
 {"query type not supported",appConnDiagProtocolError},
 {"malformed query",appConnDiagProtocolError},
 {"database unavailable",appConnDiagProtocolError},
 {"operator unsupported",appConnDiagProtocolError},
 {"too many databases specified",appConnDiagProtocolError},
 {"too many result sets created",appConnDiagProtocolError},
 {"unsupported attribute type",appConnDiagProtocolError},
 {"unsupported use attribute",appConnDiagProtocolError},
 {"unsupported value for use attribute",appConnDiagProtocolError},
 {"use attribute required but not supplied",appConnDiagProtocolError},
 {"unsupported relation attribute",appConnDiagProtocolError},
 {"unsupported structure attribute",appConnDiagProtocolError},
 {"unsupported position attribute",appConnDiagProtocolError},
 {"unsupported truncation attribute",appConnDiagProtocolError},
 {"unsupported attribute set",appConnDiagProtocolError},
 {"unsupported completeness attribute",appConnDiagProtocolError},
 {"unsupported attribute combination",appConnDiagProtocolError},
 {"unsupported coded value for term",appConnDiagProtocolError},
 {"malformed search term",appConnDiagProtocolError},
 {"illegal term value for attribute",appConnDiagProtocolError},
 {"unparseable format for un-normalized value",appConnDiagProtocolError},
 {"illegal result set name",appConnDiagProtocolError},
 {"proximity search of sets not supported",appConnDiagProtocolError},
 {"illegal result set in proximity search",appConnDiagProtocolError},
 {"unsupported proximity relation",appConnDiagProtocolError},
 {"unsupported proximity unit code",appConnDiagProtocolError},

/* diagnostic 200-231 */
 {"unknown condition",appConnDiagProtocolError},
 {"proximity not supported with this attribute combination",appConnDiagProtocolError},
 {"unsupported distance for proximity",appConnDiagProtocolError},
 {"ordered flag not supported for proximity",appConnDiagProtocolError},
 {"exclusion flag not supported for proximity",appConnDiagProtocolError},
 {"only zero step size supported for scan",appConnDiagProtocolError},
 {"specified step size not supported for scan",appConnDiagProtocolError},
 {"cannot sort according to sequence",appConnDiagProtocolError},
 {"no result set name supplied on sort",appConnDiagProtocolError},
 {"generic sort not supported (db specific only)",appConnDiagProtocolError},
 {"db specific sort not supported",appConnDiagProtocolError},
 {"too many sort keys",appConnDiagProtocolError},
 {"duplicate sort keys",appConnDiagProtocolError},
 {"unsupported missing data action",appConnDiagProtocolError},
 {"illegal sort relation",appConnDiagProtocolError},
 {"illegal case value",appConnDiagProtocolError},
 {"illegal missing data action",appConnDiagProtocolError},
 {"cannot guarantee records will fit in specified segments",appConnDiagProtocolError},
 {"extended services: package name already in use",appConnDiagProtocolError},
 {"extended services: no such package, on modify/delete",appConnDiagProtocolError},
 {"extended services: quota exceeded",appConnDiagProtocolError},
 {"extended services: extended service type not supported",appConnDiagProtocolError},
 {"extended services: permission denied on extended services - id not authorized",appConnDiagProtocolError},
 {"extended services: permission denied on extended services - cannot modify or delete",appConnDiagProtocolError},
 {"extended services: immediate execution failed",appConnDiagProtocolError},
 {"extended services: immediate execution not supported for this service",appConnDiagProtocolError},
 {"extended services: immediate execution not supported for these parameters",appConnDiagProtocolError},
 {"no data available in requested record syntax",appConnDiagProtocolError},
 {"scan: malformed scan",appConnDiagProtocolError},
 {"term type not supported",appConnDiagProtocolError},
 {"sort: too many input results",appConnDiagProtocolError},
 {"sort: incompatible record formats",appConnDiagProtocolError},

 };



long
diagCondEntry(long pCond)
{
  long lOffset = 0L;
  
  if ((pCond >= DIAG_SET_BIB1_000_START) && 
      (pCond < DIAG_SET_BIB1_000_START + DIAG_SET_BIB1_000_LEN))

    /* If it is a recogized condition from the 0-100 Diag1 Set. */
    /* Calculate correct offset in diag1 table. */
    lOffset = pCond;

  else if ((pCond >= DIAG_SET_BIB1_100_START) && 
	   (pCond < DIAG_SET_BIB1_100_START + DIAG_SET_BIB1_100_LEN))

    /* If it is a recognized condition from the 100-200 Diag1 Set. */
    /* Calculate correct offset in diag1 table. */
    lOffset = pCond - DIAG_SET_BIB1_100_START + DIAG_SET_BIB1_000_LEN;

  else if ((pCond >= DIAG_SET_BIB1_200_START) && 
	   (pCond < DIAG_SET_BIB1_200_START + DIAG_SET_BIB1_200_LEN))

    /* If it is a recogized condition from the 200-300 Diag1 Set. */
    /* Calculate correct offset in diag1 table. */
    lOffset = pCond - DIAG_SET_BIB1_200_START + 
      DIAG_SET_BIB1_000_LEN + DIAG_SET_BIB1_100_LEN;

  return(lOffset);
}



char*
diagCondMeaning(long pCond)
{
  long lOffset = diagCondEntry(pCond);
  return(gBib1Diag[lOffset].sZ3950V2Info);
}



long
diagCondAppConnDiag(long pCond)
{
  long lOffset = diagCondEntry(pCond);
  return(gBib1Diag[lOffset].sAppConnDiagCond);
}

