/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Module:   tag.c

 Purpose:  This module contains utility functions for dumping BER encodings
           to a file, and for searching tags in a directory tree.
---------------------------------------------------------------------------*/

#include <stddef.h>
#include <stdio.h>
#include "berutil.h"
#include "tag.h"

/* For file dumps of the BER */
static FILE *gLogFile = NULL; 

char *gLogFileName = "LogFile";



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:         BufDumpBer
 Purpose:          Utility function for dumping the hex ber to a file.
 Parameters:       pBerName -- input; null-terminated string that describes
                               the packet being dumped to file.
                   pDir     -- input; the directory tree to be dumped.
 Global Variables: gLogFile, gLogFileName
 Returns:          void
---------------------------------------------------------------------------*/
void 
BufDumpBer(char *pBerName, DATA_DIR *pDir) {
  if (gLogFile == NULL) {
     gLogFile=fopen(gLogFileName,"w");
     fclose(gLogFile);
  }
  gLogFile=fopen(gLogFileName,"a");
  fprintf(gLogFile,"%s length = %ld\n",pBerName,rec_len(pDir));
  hex_dirf(pDir, 0, gLogFile);
  fclose(gLogFile);
  return;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:         tagFindNext
 Purpose:          Utility function.  Given a data directory structure, finds
                   and returns the next directory in the sequence that contains
                   the specified tag.
 Parameters:       pStart -- input; a handle to a DATA_DIR directory tree.
                   pTag   -- input; the tag being searched for.  Note that
                   berutil expects fldid to be a short unsigned.
 Global Variables: None.
 Returns:          A handle to a DATA_DIR representing the next DATA_DIR in
                   the sequence that contains the specified tag.
---------------------------------------------------------------------------*/
DATA_DIR 
*tagFindNext(DATA_DIR *pStart, unsigned pTag)
{
    DATA_DIR *lNode = pStart;
    while (lNode) {
      if (lNode->fldid == pTag) return lNode;
      lNode = lNode->next;
    }
    return (NULL);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:         tagFindChild
 Purpose:          Utility function.  Given a data directory structure, find
                   and return the child directory containing the specified tag.
 Parameters:       pStart -- input; a handle to a DATA_DIR directory tree.
                   pTag   -- input; the tag being search for.  Note that
                   berutil expects fldid to be a short unsigned.
 Global Variables: None.
 Returns:          A handle to the DATA_DIR representing the child DATA_DIR
                   containing the specified tag.
---------------------------------------------------------------------------*/
DATA_DIR 
*tagFindChild(DATA_DIR *pStart, unsigned pTag)
{
    DATA_DIR *lNode = pStart->ptr.child;
    while (lNode) {
      if (lNode->fldid == pTag) return lNode;
      lNode = lNode->next;
    }
    return (NULL);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:         dreplace_num
 Purpose:          Utility function.  Replaces an number in a directory tree
                   pDir with the number given by pnum.  It isolates the caller
                   from the underlying data structure of pDir.  This function
                   should have been specified in the berutil.c file, but was
                   not included in the OCLC release.
 Parameters:       pDir -- input; a handle to a DATA_DIR directory tree.
                   pNum -- input; the new number.
 Global Variables: None.
 Returns:          A handle to a DATA_DIR containing the new number.
---------------------------------------------------------------------------*/
DATA_DIR
*dreplace_num(DATA_DIR *pDir,long pNum)
{
  if (pDir != NULL)
    { pDir->number = pNum;
      return(pDir);
    }
  else return(NULL);
}
