/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

#ifndef tag_h
#define tag_h

/* ASN.1 universal class tag assignments                               */
#define tagUnivBool         1L     /* Boolean type                     */
#define tagUnivInt          2L     /* Integer type                     */
#define tagUnivBitStr       3L     /* Bitstring type                   */
#define tagUnivOctStr       4L     /* Octetstring type                 */
#define tagUnivNull         5L     /* Null type                        */
#define tagUnivObjId        6L     /* Object identifier type           */
#define tagUnivExt          8L     /* External type                    */
#define tagUnivEnum        10L     /* Enumerated type                  */
#define tagUnivSeq         16L     /* Sequence and sequence-of types   */
#define tagUnivSet         17L     /* Set and set-of types             */
#define tagUnivGenTime     24L     /* Generalized Time                 */
#define tagUnivVisStr      26L     /* Visible string type              */
#define tagUnivGenStr      27L     /* General string                   */

/* The PDU tags used globally.                                         */
#define tagGlobRefId        2L     /* Reference identifier             */
#define tagGlobDbName     105L     /* Database name                    */
#define tagGlobResSetId    31L     /* Result set identifier            */

/* The PDU tags for Z39.50 Version 2.                                   */
#define tagPDUInitReq        20L   /* Initialize request                */
#define tagPDUInitResp       21L   /* Initialize response               */
#define tagPDUSrchReq        22L   /* Search request                    */
#define tagPDUSrchResp       23L   /* Search response                   */
#define tagPDUPrsntReq       24L   /* Present request                   */
#define tagPDUPrsntResp      25L   /* Present response                  */
#define tagPDUDlteResSetReq  26L   /* Delete result set request         */
#define tagPDUDlteResSetResp 27L   /* Delete result set response        */
#define tagPDUAccCtrlReq     28L   /* Access control request            */
#define tagPDUAccCtrlResp    29L   /* Access control response           */
#define tagPDURscCtrlReq     30L   /* Resource control request          */
#define tagPDURscCtrlResp    31L   /* Resource control response         */
#define tagPDUTrgRscCtrlReq  32L   /* Trigger resource control request  */
#define tagPDURscRptReq      33L   /* Resource report request           */
#define tagPDURscRptResp     34L   /* Resource report response          */

/* The PDU tags for Initialize Request and Initialize Response.     */
#define tagInitProtVer        3L   /* Protocol version to be used   */
#define tagInitOpts           4L   /* Set of services to be used    */
#define tagInitPrefMsgSiz     5L   /* Preferred message size        */
#define tagInitMaxRecSiz      6L   /* Maximum record size           */
#define tagInitIdAuth         7L   /* Identification authentication */
#define tagInitUsrInfoFld    11L   /* User information field        */
#define tagInitRes           12L   /* Result of the processing of   */
                                   /* the request at the target     */
#define tagInitImpId        110L   /* Implementation identifier     */
#define tagInitImpName      111L   /* Implementation name           */
#define tagInitImpVer       112L   /* Implementation version        */

/* The PDU tags used for both Search and Present.                      */
#define tagSrchAndPrsntNumRecRet    24L /* Number of Records Returned  */
#define tagSrchAndPrsntNxtResSetPos 25L /* Next Result Set Position    */
#define tagSrchAndPrsntPrsntStat    27L /* Present Status              */
#define tagSrchAndPrsntPrefRecSyn  104L /* Preferred Record Syntax     */

/* The PDU Tags for Search Request and Search Response.                  */
#define tagSrchSmlSetUpBnd         13L  /* Small Set Upper Bound         */
#define tagSrchLrgSetLoBnd         14L  /* Large Set Lower Bound         */
#define tagSrchMedSetPrsntNum      15L  /* Medium Set Present Number     */
#define tagSrchReplInd             16L  /* Replacement Indicator         */
#define tagSrchResSetName          17L  /* Result Set Name               */
#define tagSrchDbNames             18L  /* Database Names                */
#define tagSrchSmlSetElmtSetName  100L  /* Small Set Element Set Names   */
#define tagSrchMedSetElmtSetName  101L  /* Medium Set Element Set Names  */
#define tagSrchQry                 21L  /* Query                         */
#define tagSrchResCnt              23L  /* Result Count                  */
#define tagSrchSrchStat            22L  /* Search Status                 */
#define tagSrchResSetStat          26L  /* Result Set Status             */

/* The PDU Tags for Present Request and Present Response                 */
#define tagPrsntElmtSetNames      19L     /* Element Set Name            */
#define tagPrsntResSetStrtPt      30L     /* Result Set Start Point      */
#define tagPrsntNumRecReq         29L     /* Number of Records Requested */

#define tagRPNQuery          1L  /* Type-1 RPN Query               */
#define tagRPNOperand        0L  /* RPNStructure Operand           */
#define tagRPNSeq            1L  /* RPNStructure Implicit Sequence */
#define tagRPNAttrList      44L  /* Attribute List                 */
#define tagRPNTerm          45L  /* Term                           */
#define tagRPNOperator      46L  /* Operator                       */
#define tagRPNAttrPlusTerm 102L  /* Attribute Plus Term            */
#define tagRPNAttrType     120L  /* Attribute Type                 */
#define tagRPNAttrVal      121L  /* Attribute Value                */
#define tagRPNOperatorAnd    0L  /* RPN Operator AND               */
#define tagRPNOperatorOr     1L  /* RPN Operator Or                */
#define tagRPNOperatorAndNot 2L  /* RPN Operator AndNot            */

#define tagElmtSetNamesGen   0L  /* Element Set Names Generic      */
#define tagElmtSetNamesSpec  1L  /* Element Set Names Specific     */
#define tagElmtSetName     103L  /* Element Set Name               */

#define tagResSetStatSub 1L      /* Subset */
#define tagResSetStatInt 2L      /* Interum */
#define tagResSetStatNon 3L      /* None */

#define tagPrsntStatSucc   0L       /* Success */
#define tagPrsntStatPart1  1L       /* Partial-1 */
#define tagPrsntStatPart2  2L       /* Partial-2 */
#define tagPrsntStatPart3  3L       /* Partial-3 */ 
#define tagPrsntStatPart4  4L       /* Partial-4 */ 
#define tagPrsntStatFail   5L       /* Failure */

#define tagRecDbOrSurDiag                28L
#define tagRecNonSurDiag                130L
#define tagRecNamePlusRecDbName           0L
#define tagRecNamePlusRecDbRecOrSurDiag   1L
#define tagRecNamePlusRecDbRec            1L
#define tagRecNamePlusRecDbSurDiag        2L

#define tagExtEncSin       0L /* any, single ASN1-type */
#define tagExtEncOctAlign  1L /* implicit octet string, octet-aligned */
#define tagExtEncArb       2L /* implicit bit string, arbitrary */

#define tagTagSetM     1L /* Meta-Information 1.2.840.10003.14.1 */
#define tagTagSetG     2L /* Generic-Information 1.2.840.10003.14.2 */
#define tagTagSetL     3L /* Locally-Defined Tags */

/* Generic Information TagSet */
#define tagSetGHeadline  1L
#define tagSetGName      2L
#define tagSetGDate      3L

/* Meta Information TagSet */
#define tagSetMRank     10L
#define tagSetMScore    18L
#define tagSetMRecId    14L
#define tagSetMURx      12L

#define tagGRSTagType      1L /* tag type */
#define tagGRSTagVal       2L /* tag value */
#define tagGRSContent      4L /* element data */
#define tagGRSMetaData     5L /* element metadata */

#define tagGRSTagValStr    1L /* String Tag Value */
#define tagGRSTagValNum    2L /* Numeric Tag Value */

#define tagGRSElmtDataIntUnit         1L
#define tagGRSElmtDataElmtNotThere    2L
#define tagGRSElmtDataElmtEmpty       3L
#define tagGRSElmtDataNoDataRequested 4L
#define tagGRSElmtDataDiag            5L
#define tagGRSElmtDataSubTree         6L

#define tagGRSElmtMetaDataSeriesOrder 1L
#define tagGRSElmtMetaDataUsageRight  2L
#define tagGRSElmtMetaDataHits        3L
#define tagGRSElmtMetaDataDisplayName 4L
#define tagGRSElmtMetaDataSupVariants 5L

#define tagGRSHitVecSatisfier         1L
#define tagGRSHitVecOffsetIntoElement 2L
#define tagGRSHitVecLength            3L
#define tagGRSHitVecHitRank           4L
#define tagGRSHitVecTargetToken       5L

#define tagGRSVarGlblVarSetId 1L
#define tagGRSVarTriples      2L

#define tagGRSVarTriplesVarId 0L
#define tagGRSVarTriplesClass 1L
#define tagGRSVarTriplesType  2L
#define tagGRSVarTriplesValue 3L

/* Variant-1 */
#define tagVarClassVarId        1L
#define tagVarClassBodyPartType 2L
#define tagVarClassFmtPres     3L
#define tagVarClassLangCharSet 4L
#define tagVarClassPiece        5L
#define tagVarClassMetaDataReq  6L
#define tagVarClassMedaDataRet  7L
#define tagVarClassMisc         8L

#define tagVarClassVarIdVarId   1L

#define tagVarClassBodyPartTypeIana      1L
#define tagVarClassBodyPartTypeZ3950Type 2L
#define tagVarClassBodyPartTypeOtherType 3L

#define tagVarClassLangCharSetLang        1L
#define tagVarClassLangCharSetRegCharSet  2L
#define tagVarClassLangCharSetPrivCharSet 3L

#define tagVarClassPieceFragWanted  1L
#define tagVarClassPieceFragRet     2L
#define tagVarClassPieceStart       3L
#define tagVarClassPieceEnd         4L
#define tagVarClassPieceHowMuch     5L
#define tagVarClassPieceStep        6L
#define tagVarClassPieceTargetToken 7L

#define tagVarClassMedaDataRetCost  1L 
#define tagVarClassMedaDataRetSize  2L
#define tagVarClassMedaDataRetInteg 3L


/* Tag utilities. */
void BufDumpBer(char *pBerName, DATA_DIR *pDir);

DATA_DIR *tagFindNext(DATA_DIR* pStart, 
		      unsigned  pTag);

DATA_DIR *tagFindChild(DATA_DIR* pStart, 
		       unsigned  pTag);

#endif /* tag_h */
